# -*- coding: utf-8 -*-
#!/usr/bin/python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007      Thom Sturgill
# Copyright (C) 2007-2009 Brian G. Matherly
# Copyright (C) 2008-2011 Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2008      Jason Simanek
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2015-     Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Web Calendar generator.
"""

# ------------------------------------------------------------------------
# python modules
# ------------------------------------------------------------------------
import os
import shutil
import datetime
import time
import calendar  # Python module

# ------------------------------------------------------------------------
# Set up logging
# ------------------------------------------------------------------------
import logging

# ------------------------------------------------------------------------
# Gramps module
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import Date, Name, NameType, Person
from gramps.gen.lib.date import Today
from gramps.gen.const import PROGRAM_NAME, URL_HOMEPAGE
from gramps.version import VERSION
from gramps.gen.constfunc import win
from gramps.gen.config import config
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.plug.menu import (
    BooleanOption,
    NumberOption,
    StringOption,
    EnumeratedListOption,
    FilterOption,
    PersonOption,
    DestinationOption,
    NoteOption,
)
from gramps.gen.utils.config import get_researcher
from gramps.gen.utils.alive import probably_alive
from gramps.gen.utils.db import get_death_or_fallback
from gramps.gen.utils.symbols import Symbols
from gramps.gen.datehandler import displayer as _dd

from gramps.gen.display.name import displayer as _nd

import gramps.plugins.lib.libholiday as libholiday
from gramps.plugins.lib.libhtml import Html, xml_lang
from gramps.plugins.lib.libhtmlconst import _CHARACTER_SETS, _CC, _COPY_OPTIONS
from gramps.gui.pluginmanager import GuiPluginManager
from gramps.plugins.webreport.common import html_escape

from gramps.gen.lib.date import gregorian

# import styled notes from
# src/plugins/lib/libhtmlbackend.py
from gramps.plugins.lib.libhtmlbackend import HtmlBackend

# ------------------------------------------------------------------------
# constants
# ------------------------------------------------------------------------
_ = glocale.translation.sgettext
_LOG = logging.getLogger(".WebPage")

# full clear line for proper styling

FULLCLEAR = Html("div", class_="fullclear", inline=True)

# Web page filename extensions
_WEB_EXT = [".html", ".htm", ".shtml", ".php", ".php3", ".cgi"]

# Calendar stylesheet names
_CALENDARSCREEN = "calendar-screen.css"
_CALENDARPRINT = "calendar-print.css"

PLUGMAN = GuiPluginManager.get_instance()
CSS = PLUGMAN.process_plugin_data("WEBSTUFF")


# ------------------------------------------------------------------------
#
# WebCalReport
#
# ------------------------------------------------------------------------
class WebCalReport(Report):
    """
    Create WebCalReport object that produces the report.
    """

    def __init__(self, database, options, user):
        Report.__init__(self, database, options, user)
        self._user = user

        stdoptions.run_private_data_option(self, options.menu)

        # class to do conversion of styled notes to html markup
        self._backend = HtmlBackend()

        self.options = options

        mgobn = lambda name: options.menu.get_option_by_name(name).get_value()

        self.set_locale(options.menu.get_option_by_name("trans").get_value())
        stdoptions.run_date_format_option(self, options.menu)
        self.rlocale = self._locale
        self._ = self.rlocale.translation.sgettext

        self.html_dir = mgobn("target")
        self.title_text = html_escape(mgobn("title"))
        filter_option = options.menu.get_option_by_name("filter")
        self.filter = filter_option.get_filter()
        self.name_format = mgobn("name_format")
        self.ext = mgobn("ext")
        self.copy = mgobn("cright")
        self.css = mgobn("css")

        self.country = mgobn("country")
        self.start_dow = mgobn("start_dow")

        self.multiyear = mgobn("multiyear")
        self.start_year = mgobn("start_year")
        self.end_year = mgobn("end_year")
        self.after_year = mgobn("after_year")
        if not self.multiyear:
            self.end_year = self.start_year
        if self.end_year < self.start_year:
            self.end_year = self.start_year

        self.maiden_name = mgobn("maiden_name")

        self.alive = mgobn("alive")
        self.birthday = mgobn("birthdays")
        self.anniv = mgobn("anniversaries")
        self.death_anniv = mgobn("death_anniv")
        self.event_list = []

        self.month_notes = [
            mgobn("note_" + month)
            for month in [
                "jan",
                "feb",
                "mar",
                "apr",
                "may",
                "jun",
                "jul",
                "aug",
                "sep",
                "oct",
                "nov",
                "dec",
            ]
        ]

        self.encoding = mgobn("encoding")
        self.fullyear = True
        self.makeoneday = mgobn("makeoneday")

        # identify researcher name and e-mail address
        # as NarrativeWeb already does
        researcher = get_researcher()
        self.author = researcher.name
        if self.author:
            self.author = self.author.replace(",,,", "")
        self.email = researcher.email

        # set to today's date
        self.today = Today()

        self.warn_dir = True  # Only give warning once.

        self.link_to_narweb = mgobn("link_to_narweb")
        self.narweb_prefix = mgobn("prefix")

        # self.calendar is a dict; key is the month number
        # Each entry in the dict is also a dict; key is the day number.
        # The day dict is a list of things to display for that day.
        # These things are: birthdays and anniversaries
        self.calendar = {}
        self.holidays = {}

        calendar.setfirstweekday(DOW_GRAMPS2ISO[self.start_dow])
        self.head = []
        # An optional link to a home page
        self.head.append(
            (self.narweb_prefix, self._("NarrativeWeb Home"), self.link_to_narweb)
        )

    def get_note_format(self, note):
        """
        will get the note from the database, and will return either the
        styled text or plain note
        """

        # retrieve the body of the note
        note_text = note.get()

        # styled notes
        htmlnotetext = self.styled_note(note.get_styledtext(), note.get_format())
        text = htmlnotetext or Html("p", note_text)

        # return text of the note to its callers
        return text

    #################################################
    # Will produce styled notes for WebCal by using:
    # src/plugins/lib/libhtmlbackend.py
    #################################################
    def styled_note(self, styledtext, format_type):
        """
        styledtext : assumed a StyledText object to write
        format_type : = 0 : Flowed, = 1 : Preformatted
        style_name : name of the style to use for default presentation
        """
        text = str(styledtext)

        if not text:
            return ""

        s_tags = styledtext.get_tags()
        # FIXME: following split should be regex to match \n\s*\n instead?
        markuptext = self._backend.add_markup_from_styled(text, s_tags, split="\n\n")
        htmllist = Html("div", id="grampsstylednote")
        if format_type == 1:
            # preformatted, retain whitespace.
            # so use \n\n for paragraph detection
            # FIXME: following split should be regex to match \n\s*\n instead?
            htmllist += Html("pre", indent=None, inline=True)
            for line in markuptext.split("\n\n"):
                htmllist += Html("p")
                for realline in line.split("\n"):
                    htmllist += realline
                    htmllist += Html("br")

        elif format_type == 0:
            # flowed
            # FIXME: following split should be regex to match \n\s*\n instead?
            for line in markuptext.split("\n\n"):
                htmllist += Html("p")
                htmllist += line

        return htmllist

    def copy_file(self, from_fname, to_fname, to_dir=""):
        """
        Copy a file from a source to a (report) destination.
        If to_dir is not present and if the target is not an archive,
        then the destination directory will be created.

        Normally 'to_fname' will be just a filename, without directory path.

        'to_dir' is the relative path name in the destination root. It will
        be prepended before 'to_fname'.
        """
        dest = os.path.join(self.html_dir, to_dir, to_fname)

        destdir = os.path.dirname(dest)
        if not os.path.isdir(destdir):
            os.makedirs(destdir)

        if from_fname != dest:
            shutil.copyfile(from_fname, dest)
        elif self.warn_dir:
            self._user.warn(
                _("Possible destination error")
                + "\n"
                + _(
                    "You appear to have set your target directory "
                    "to a directory used for data storage. This "
                    "could create problems with file management. "
                    "It is recommended that you consider using "
                    "a different directory to store your generated "
                    "web pages."
                )
            )
            self.warn_dir = False
        config.set("paths.website-directory", os.path.dirname(self.html_dir) + os.sep)

    def add_day_item(
        self, text, year, month, day, event, age_at_death, dead_event_date
    ):
        """
        adds birthdays, anniversaries, and holidays to their perspective lists

        text -- line to be added
        year, month, day -- date to add the text to

        event -- one of 'BirthDay', 'Anniversary', 'Death' or 'Holiday'
        age_at_death -- The age in text. ie : 68 years, 6 months
        dead_event_date -- The date of the event used to calculate
                           the age_at_death
        """

        if year <= self.after_year:
            return

        # This may happen for certain "about" dates.
        # Use first day of the month
        if day == 0:
            day = 1

        # determine which dictionary to use???
        if event in ["Birthday", "Anniversary", "Death"]:
            month_dict = self.calendar.get(month, {})
        else:
            month_dict = self.holidays.get(month, {})
        day_list = month_dict.get(day, [])

        if month > 0:
            try:
                event_date = Date(year, month, day)
            except ValueError:
                event_date = Date.EMPTY
        else:
            event_date = Date.EMPTY  # Incomplete date....

        day_list.append((text, event, event_date, age_at_death, dead_event_date))
        month_dict[day] = day_list

        # determine which dictionary to add it to???
        if event in ["Birthday", "Anniversary", "Death"]:
            self.calendar[month] = month_dict
        else:
            self.holidays[month] = month_dict

    def __get_holidays(self, year):
        """Get the holidays for the specified country and year"""

        # _('translation')
        with self._user.progress(
            _("Web Calendar Report"),
            _("Calculating Holidays for year %04d") % year,
            365,
        ) as step:
            holiday_table = libholiday.HolidayTable()
            country = holiday_table.get_countries()[self.country]
            holiday_table.load_holidays(year, country)
            for month in range(1, 13):
                for day in range(1, 32):
                    holiday_names = holiday_table.get_holidays(month, day)
                    for holiday_name in holiday_names:
                        self.add_day_item(
                            holiday_name, year, month, day, "Holiday", None, None
                        )
                    step()

    def copy_calendar_files(self):
        """
        Copies all the necessary stylesheets and images for these calendars
        """
        imgs = []

        # copy all screen style sheet
        for css_f in CSS:
            already_done = False
            for css_fn in ("UsEr_", "Basic", "Mainz", "Nebraska", "Vis"):
                if css_fn in css_f and not already_done:
                    already_done = True
                    fname = CSS[css_f]["filename"]
                    # add images for this css
                    imgs += CSS[css_f]["images"]
                    css_f = css_f.replace("UsEr_", "")
                    self.copy_file(fname, css_f + ".css", "css")

        # Copy the screen stylesheet
        if self.css and self.css != "No style sheet":
            fname = CSS[self.css]["filename"]
            self.copy_file(fname, _CALENDARSCREEN, "css")

        # copy Navigation Menu Layout if Blue or Visually is being used
        if CSS[self.css]["navigation"]:
            # copy horizontal menus...
            fname = CSS["Horizontal-Menus"]["filename"]
            self.copy_file(fname, "calendar-menus.css", "css")

        # copy print stylesheet
        fname = CSS["Print-Default"]["filename"]
        self.copy_file(fname, _CALENDARPRINT, "css")

        # Mainz stylesheet graphics
        # will only be used if Mainz is slected as the stylesheet
        imgs += CSS[self.css]["images"]

        # copy copyright image
        # the proper way would be to call "filename", but it is NOT working...
        if 0 < self.copy <= len(_CC):
            imgs += [CSS["Copyright"]["filename"]]

        # copy Gramps favicon #2
        imgs += [CSS["favicon2"]["filename"]]

        for from_path in imgs:
            dummy_fdir, fname = os.path.split(from_path)
            self.copy_file(from_path, fname, "images")

    def create_file(self, fname, subdir):
        """
        Create a file in the html_dir tree.
        If the directory does not exist, create it.

        fname -- filename to be created
        subdir -- any subdirs to be added
        """

        fname = os.path.join(self.html_dir, subdir, fname)

        if not _has_webpage_extension(fname):
            fname += self.ext

        destdir = os.path.dirname(fname)

        if not os.path.isdir(destdir):
            os.makedirs(destdir)

        output_file = open(
            fname, "w", encoding=self.encoding, errors="xmlcharrefreplace"
        )
        return output_file

    def close_file(self, output_file):
        """will close whatever filename is passed to it"""
        output_file.close()

    def write_header(self, nr_up, title, body_id=None, add_print=True):
        """
        This creates the header for the Calendars
        'nr_up' - number of directory levels up, started from current page, to
                  the root of the directory tree (i.e. to self.html_dir).
        title -- to be inserted into page header section
        add_print -- whether to add printer stylesheet or not
            * only webcalendar() and one_day() only!
        """

        # number of subdirectories up to reach root
        subdirs = [".."] * nr_up

        # Header contants
        xmllang = xml_lang()
        _meta1 = (
            'name ="viewport" content="width=device-width, '
            'initial-scale=1.0, maximum-scale=1.0, user-scalable=1"'
        )
        _meta2 = 'name ="apple-mobile-web-app-capable" content="yes"'
        _meta3 = 'name="generator" content="%s %s %s"' % (
            PROGRAM_NAME,
            VERSION,
            URL_HOMEPAGE,
        )
        _meta4 = 'name="author" content="%s"' % self.author

        # create additional meta tags
        meta = Html("meta", attr=_meta1) + (
            Html("meta", attr=_meta2, indent=False),
            Html("meta", attr=_meta3, indent=False),
            Html("meta", attr=_meta4, indent=False),
        )

        # begin each html page...
        page, head, body = Html.page(title, self.encoding, xmllang)

        # Add the script to control the menu
        menuscript = Html(
            "<script>function navFunction() { "
            'var x = document.getElementById("dropmenu"); '
            'if (x.className === "nav") { x.className += "'
            ' responsive"; } else { x.className = "nav"; }'
            " }</script>"
        )
        head += menuscript

        # begin header section
        headerdiv = (
            Html("div", id="header")
            + (
                Html(
                    '<button href="javascript:void(0);" class="navIcon"'
                    ' onclick="navFunction()">&#8801;</button>'
                )
            )
            + (Html("h1", self.title_text, id="SiteTitle", inline=True))
        )
        body += headerdiv

        # add body id tag if not None
        if body_id is not None:
            body.attr = "id = '%(idtag)s'" % {"idtag": body_id}

        # Gramps favicon
        fname1 = "/".join(subdirs + ["images", "favicon2.ico"])

        # _CALENDARSCREEN stylesheet
        fname2 = "/".join(subdirs + ["css", _CALENDARSCREEN])

        # links for Gramps favicon and stylesheets
        links = Html("link", rel="shortcut icon", href=fname1, type="image/x-icon") + (
            Html(
                "link",
                href=fname2,
                type="text/css",
                title=self._("Default"),
                media="screen",
                rel="stylesheet",
                indent=False,
            )
        )
        # create all alternate stylesheets
        # Cannot use it on local files (file://)
        for css_f in CSS:
            already_done = False
            for css_fn in ("UsEr_", "Basic", "Mainz", "Nebraska"):
                if css_fn in css_f and not already_done:
                    css_f = css_f.replace("UsEr_", "")
                    fname = "/".join(subdirs + ["css", css_f + ".css"])
                    links += Html(
                        "link",
                        rel="alternate stylesheet",
                        title=css_f,
                        indent=False,
                        media="screen",
                        type="text/css",
                        href=fname,
                    )

        # add horizontal menu if css == Blue or Visually because
        # there is no menus?
        if CSS[self.css]["navigation"]:
            fname = "/".join(subdirs + ["css", "calendar-menus.css"])
            links.extend(
                Html(
                    "link",
                    href=fname,
                    type="text/css",
                    media="screen",
                    rel="stylesheet",
                    indent=False,
                )
            )

        # add printer stylesheet to webcalendar() and one_day() only
        if add_print:
            fname = "/".join(subdirs + ["css", _CALENDARPRINT])
            links.extend(
                Html(
                    "link",
                    href=fname,
                    type="text/css",
                    media="print",
                    rel="stylesheet",
                    indent=False,
                )
            )

        # add meta tags and links to head section
        head += (meta, links)

        # start header section and page title...
        script = (
            """
<script type="text/javascript">
function currentmonth(y) {
var date = new Date();
var month = date.getMonth() + 1;
var url = y + "/" + month + "%s";
window.location.href = url;
return false;
}
</script>
"""
            % self.ext
        )
        body += script
        return page, body

    def year_navigation(self, nr_up, currentsection):
        """
        This will create the year navigation menu bar for a total of
        seventeen (17) years

        nr_up = number of directories up to reach root directory
        currentsection = proper styling of this navigation bar
        """

        # limit number of years to eighteen (18) years and only one row of years
        nyears = (self.end_year - self.start_year) + 1
        num_years = nyears if 0 < nyears < 19 else 18
        self.end_year = (self.start_year + 17) if nyears > 18 else self.end_year

        # begin year division and begin unordered list
        with Html(
            "div", class_="wrappernav", id="nav", role="navigation"
        ) as navigation:
            with Html("div", class_="container") as container:
                unordered = Html("ul", class_="nav", id="dropmenu")

                (url, nav_text, disp) = self.head[0]
                if disp:
                    if url[:1] == "/":
                        url = url + "index" + self.ext
                    else:
                        url_up = [".."] * nr_up
                        url_up.append(url)
                        url = "/".join(url_up) + "index" + self.ext
                    hyper = Html("a", nav_text, href=url, name=url, title=nav_text)
                    unordered.extend(Html("li", hyper, inline=True))

                for cal_year in range(self.start_year, (self.start_year + num_years)):
                    url = ""

                    # begin subdir level
                    subdirs = [".."] * nr_up
                    subdirs.append(str(cal_year))

                    # Note. We use '/' here because it is a URL,
                    # not a OS dependent pathname.
                    url = "/".join(subdirs)
                    onclic = "return currentmonth('" + url + "');"
                    hyper = Html(
                        "a",
                        self.rlocale.get_date(Date(cal_year)),
                        href="#",
                        onclick=onclic,
                        title=str(cal_year),
                    )

                    # Figure out if we need <li class="CurrentSection">
                    # or just plain <li>
                    if str(cal_year) == currentsection:
                        check_cs = 'class = "CurrentSection"'
                    else:
                        check_cs = False
                    if check_cs:
                        unordered.extend(Html("li", hyper, attr=check_cs, inline=True))
                    else:
                        unordered.extend(Html("li", hyper, inline=True))
                container += unordered
            navigation += container
        return navigation

    def month_navigation(self, nr_up, year, currentsection):
        """
        Will create and display the navigation menu bar

        nr_up = number of directories up to reach root directory
        year = year being created
        currentsection = month name being created for proper CSS styling
        """
        navs = []

        if not self.multiyear:
            (url, nav_text, disp) = self.head[0]
            if url[:1] == "/":
                url = url + "index" + self.ext
            else:
                url = "../" + url + "index" + self.ext
            navs.append((url, nav_text, disp))

        navs.extend(
            (str(month), self.rlocale.date_displayer.short_months[int(month)], True)
            for month in range(1, 13)
        )

        # Add a link for year_glance() if requested
        navs.append(("fullyearlinked", self._("Full year at a Glance"), self.fullyear))

        # remove menu items if they are not True
        navs = [(u, n) for u, n, c in navs if c]

        # begin month subnavigation
        with Html("div", class_="wrapper", id="nav", role="navigation") as navigation:
            with Html("div", class_="container") as container:
                unordered = Html("ul", class_="menu")

                for url_fname, nav_text in navs:
                    # Note. We use '/' here because it is a URL, not a OS
                    # dependent pathname need to leave home link alone,
                    # so look for it ...
                    if nav_text != self._("NarrativeWeb Home"):
                        url_fname = url_fname.lower()
                    url = url_fname
                    add_subdirs = False
                    if not (url.startswith("http:") or url.startswith("/")):
                        add_subdirs = not any(url.endswith(ext) for ext in _WEB_EXT)

                    # whether to add subdirs or not???
                    if add_subdirs:
                        subdirs = [".."] * nr_up
                        subdirs.append(str(year))
                        url = "/".join(subdirs + [url_fname])

                    if not _has_webpage_extension(url):
                        url += self.ext

                    # Figure out if we need <li class="CurrentSection"> or
                    # just plain <li>
                    if url_fname == currentsection:
                        check_cs = 'class = "CurrentSection"'
                    else:
                        check_cs = False

                    if url_fname == "fullyearlinked":
                        mytitle = self._("Full year at a Glance")
                    else:
                        mytitle = self._(url_fname)
                    hyper = Html("a", nav_text, href=url, name=url_fname, title=mytitle)

                    if check_cs:
                        unordered.extend(Html("li", hyper, attr=check_cs, inline=True))
                    else:
                        unordered.extend(Html("li", hyper, inline=True))
                container += unordered
            navigation += container
        return navigation

    def calendar_build(self, cal, year, month, clickable=False):
        """
        This does the work of building the calendar

        @param: cal - either "yg" year_glance(), or "wc" webcalendar()
        @param: year -- year being created
        @param: month - month number 1, 2, .., 12
        """

        date_displayer = self.rlocale.date_displayer
        # define names for long and short month names
        full_month_name = date_displayer.long_months[int(month)]
        abbr_month_name = date_displayer.short_months[int(month)]

        # dow (day-of-week) uses Gramps numbering, sunday => 1, etc
        start_dow = self.start_dow
        col2day = [(x - 1) % 7 + 1 for x in range(start_dow, start_dow + 7)]

        def get_class_for_daycol(col):
            """Translate a Gramps day number into a HTMLclass"""
            day = col2day[col]
            if day == 1:
                return "weekend sunday"
            elif day == 7:
                return "weekend saturday"
            return "weekday"

        def get_name_for_daycol(col):
            """Translate a Gramps day number into a HTMLclass"""
            day = col2day[col]
            return day_names[day]

        # Note. gen.datehandler has sunday => 1, monday => 2, etc
        # We slice out the first empty element.
        day_names = date_displayer.long_days  # use self._ldd.long_days when
        # set_locale is used ...

        def __get_previous_month_day(year, month, day_col):
            if month == 1:
                prevmonth = calendar.monthcalendar((year - 1), 12)
            else:
                prevmonth = calendar.monthcalendar(year, (month - 1))
            num_weeks = len(prevmonth)
            lastweek_prevmonth = prevmonth[(num_weeks - 1)]
            previous_month_day = lastweek_prevmonth[day_col]

            # return previous month day number based on day_col
            # day_col is based on range(0 - 6)
            return previous_month_day

        def __get_next_month_day(year, month, day_col):
            if month == 12:
                nextmonth = calendar.monthcalendar((year + 1), 1)
            else:
                nextmonth = calendar.monthcalendar(year, (month + 1))
            firstweek_nextmonth = nextmonth[0]
            next_month_day = firstweek_nextmonth[day_col]

            # return next month day number based on day_col
            # day_col is based on range(0 - 6)
            return next_month_day

        # Begin calendar head. We'll use the capitalized name, because here it
        # seems appropriate for most countries.
        month_name = full_month_name.capitalize()
        th_txt = month_name
        if cal == "wc":  # webcalendar()
            if not self.multiyear:
                th_txt = "%s %s" % (month_name, self._get_date(Date(year)))  # localized

        # begin calendar table and table head
        with Html(
            "table", class_="calendar", id=month_name, role="Calendar-Grid"
        ) as table:
            thead = Html("thead")
            table += thead

            if clickable:
                name = str(month) + self.ext
                linkable = Html("a", th_txt, href=name, name=name, title=th_txt)
                trow = Html("tr") + (
                    Html("th", linkable, class_="monthName", colspan=7, inline=True)
                )
                thead += trow
            else:
                if not self.multiyear:
                    self.end_year = self.start_year
                if month > 1:
                    full_month_name = str(month - 1)
                    url = full_month_name.lower() + self.ext
                    prevm = Date(int(year), int(month - 1), 0)
                    my_title = Html(
                        "a",
                        "\u276e",
                        href=url,
                        close=True,
                        title=date_displayer.display(prevm),
                    )
                elif self.multiyear and year > self.start_year:
                    full_month_name = str(12)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(year - 1), url)
                    prevm = Date(int(year - 1), 12, 0)
                    my_title = Html(
                        "a",
                        "\u276e",
                        href=dest,
                        close=True,
                        title=date_displayer.display(prevm),
                    )
                else:
                    full_month_name = str(12)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(self.end_year), url)
                    prevy = Date(self.end_year, 12, 0)
                    my_title = Html(
                        "a",
                        "\u276e",
                        href=dest,
                        close=True,
                        title=date_displayer.display(prevy),
                    )
                my_title += Html("</a>&nbsp;" + month_name + "&nbsp;")
                if month < 12:
                    full_month_name = str(month + 1)
                    url = full_month_name.lower() + self.ext
                    nextd = Date(int(year), int(month + 1), 0)
                    my_title += Html(
                        "a",
                        "\u276f",
                        href=url,
                        close=True,
                        title=date_displayer.display(nextd),
                    )
                elif self.multiyear and year < self.end_year:
                    full_month_name = str(1)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(year + 1), url)
                    nextd = Date(int(year + 1), 1, 0)
                    my_title += Html(
                        "a",
                        "\u276f",
                        href=dest,
                        close=True,
                        title=date_displayer.display(nextd),
                    )
                else:
                    full_month_name = str(1)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(self.start_year), url)
                    nexty = Date(self.start_year, 1, 0)
                    my_title += Html(
                        "a",
                        "\u276f",
                        href=dest,
                        close=True,
                        title=date_displayer.display(nexty),
                    )
                trow = Html("tr") + (
                    Html("th", my_title, class_="monthName", colspan=7, inline=True)
                )
                thead += trow
                trow = Html("tr") + (Html("th", "", colspan=7, inline=True))
                thead += trow

            # Calendar weekday names header
            trow = Html("tr")
            thead += trow

            for day_col in range(7):
                dayclass = get_class_for_daycol(day_col)
                dayname = self._(get_name_for_daycol(day_col))
                trow += Html("th", class_=dayclass, inline=True) + (
                    Html("abbr", dayname[0], title=dayname)
                )

            # begin table body
            tbody = Html("tbody")
            table += tbody

            # get first of the month and month information
            (dummy_current_date, current_ord, monthinfo) = get_first_day_of_month(
                year, month
            )

            # begin calendar table rows, starting week0
            nweeks = len(monthinfo)
            for week_row in range(0, nweeks):
                week = monthinfo[week_row]

                # if you look this up in wikipedia, the first week
                # is called week0
                trow = Html("tr", class_="week%02d" % week_row)
                tbody += trow

                # begin calendar day column
                for day_col in range(0, 7):
                    dayclass = get_class_for_daycol(day_col)

                    # day number, can also be a zero -- a day before or
                    # after month
                    day = week[day_col]

                    # start the beginning variable for <td>, table cell
                    tcell_id = "%s%02d" % (abbr_month_name, day)

                    # add calendar date division
                    datediv = Html("div", day, class_="date", inline=True)
                    clickable = Html("div", datediv, class_="clickable", inline=True)

                    ### a day in the previous or next month ###
                    if day == 0:
                        # day in previous/ next month
                        specday = (
                            __get_previous_month_day(year, month, day_col)
                            if week_row == 0
                            else __get_next_month_day(year, month, day_col)
                        )

                        specclass = "previous " if week_row == 0 else "next "
                        specclass += dayclass

                        # continue table cell, <td>, without id tag
                        tcell = Html("td", class_=specclass, inline=True) + (
                            Html("div", specday, class_="date", inline=True)
                        )

                    # normal day number in current month
                    else:
                        thisday = datetime.date.fromordinal(current_ord)

                        # Something this month
                        if thisday.month == month:
                            holiday_list = self.holidays.get(month, {}).get(
                                thisday.day, []
                            )
                            bday_anniv_list = self.calendar.get(month, {}).get(
                                thisday.day, []
                            )

                            # date is an instance because of subtracting
                            # abilities in date.py
                            event_date = Date(thisday.year, thisday.month, thisday.day)

                            # get events for this day
                            day_list = get_day_list(
                                event_date,
                                holiday_list,
                                bday_anniv_list,
                                rlocale=self.rlocale,
                            )

                            # is there something this day?
                            if day_list:
                                hilightday = "highlight " + dayclass
                                tcell = Html("td", id=tcell_id, class_=hilightday)

                                # Year at a Glance
                                if cal == "yg":
                                    # make one day pages and hyperlink
                                    if self.makeoneday:
                                        # create yyyymmdd date string for
                                        # "One Day" calendar page filename
                                        fname_date = "%04d%02d%02d" % (year, month, day)
                                        fname_date += self.ext

                                        # create hyperlink to one_day()
                                        tcell += Html(
                                            "a", datediv, href=fname_date, inline=True
                                        )

                                        # only year_glance() needs this to
                                        # create the one_day() pages
                                        self.one_day(event_date, fname_date, day_list)

                                    # just year_glance(), but no one_day() pages
                                    else:
                                        # continue table cell, <td>,
                                        # without id tag
                                        tcell = Html(
                                            "td", class_=hilightday, inline=True
                                        ) + (
                                            # adds date division
                                            Html("div", day, class_="date", inline=True)
                                        )

                                # WebCal
                                else:
                                    # add date to table cell
                                    tcell += clickable

                                    # list the events
                                    unordered = Html("ul")
                                    clickable += unordered

                                    for (
                                        dummy_nyears,
                                        dummy_date,
                                        text,
                                        event,
                                        dummy_notused,
                                        dummy_notused,
                                    ) in day_list:
                                        unordered += Html(
                                            "li",
                                            text,
                                            inline=(
                                                False
                                                if (event == "Anniversary")
                                                else True
                                            ),
                                        )
                            # no events for this day
                            else:
                                # adds date division
                                date = Html("div", day, class_="date", inline=True)
                                # create empty day with date
                                tcell = Html("td", class_=dayclass, inline=True) + (
                                    # adds date division
                                    Html("div", date, class_="empty", inline=True)
                                )
                        # nothing for this month
                        else:
                            tcell = Html("td", class_=dayclass) + (
                                # adds date division
                                Html("div", day, class_="date", inline=True)
                            )

                    # attach table cell to table row
                    # close the day column
                    trow += tcell

                    # change day number
                    current_ord += 1

            if cal == "yg":
                for weeks in range(nweeks, 6):
                    # each calendar must have six weeks for proper styling
                    # and alignment
                    with Html("tr", class_="week%02d" % (weeks + 1)) as six_weeks:
                        tbody += six_weeks

                        for dummy_emptydays in range(7):
                            six_weeks += Html("td", class_="emptyDays", inline=True)

        # return calendar table to its callers
        return table

    def webcalendar(self, year):
        """
        This method provides information and header/ footer
        to the calendar month

        year -- year being created
        """

        # do some error correcting if needed
        if self.multiyear:
            if self.end_year < self.start_year:
                self.end_year = self.start_year

        nr_up = 1  # Number of directory levels up to get to self.html_dir / root

        with self._user.progress(
            _("Web Calendar Report"), _("Formatting months ..."), 12
        ) as step:
            for month in range(1, 13):
                cal_fname = str(month)
                open_file = self.create_file(cal_fname, str(year))

                # Add xml, doctype, meta and stylesheets
                # body has already been added to webcal  already once
                webcal, body = self.write_header(nr_up, self.title_text)

                # create Year Navigation menu
                if self.multiyear and ((self.end_year - self.start_year) > 0):
                    body += self.year_navigation(nr_up, str(year))

                # Create Month Navigation Menu
                # identify currentsection for proper highlighting
                currentsection = _dd.long_months[month]
                body += self.month_navigation(nr_up, year, currentsection)

                # build the calendar
                content = Html("div", class_="content", id="WebCal")
                body += content
                monthly_calendar = self.calendar_build("wc", year, month)
                content += monthly_calendar

                # create note section for webcalendar()
                # One has to be minused because the array starts at zero,
                # but January =1
                note = self.month_notes[month - 1].strip()
                if note:
                    note = self.database.get_note_from_gramps_id(note)
                    note = self.get_note_format(note)

                # table foot section
                cal_foot = Html("tfoot")
                monthly_calendar += cal_foot

                trow = Html("tr") + (Html("td", note, colspan=7, inline=True))
                cal_foot += trow

                # create blank line for stylesheets
                # create footer division section
                footer = self.write_footer(nr_up)
                body += (FULLCLEAR, footer)

                # send calendar page to web output
                # and close the file
                self.xhtmlwriter(webcal, open_file)

                step()

    def year_glance(self, year):
        """
        This method will create the Full Year At A Glance Page...
        year -- year being created
        """

        self.event_list = []
        prv = None
        nxt = None
        evdte = None
        for month in sorted(self.calendar):
            vals = sorted(self.calendar.get(month, {}))
            if month == 0:  # why ?
                continue
            for day in vals:
                event_date = "%04d%02d%02d" % (year, month, day)
                if evdte is None:
                    evdte = event_date
                elif nxt is None:
                    nxt = event_date
                    self.event_list.append((evdte, prv, nxt))
                else:
                    prv = evdte
                    evdte = nxt
                    nxt = event_date
                    self.event_list.append((evdte, prv, nxt))
        self.event_list.append((nxt, evdte, None))

        nr_up = 1  # Number of directory levels up to get to root

        # generate progress pass for "Year At A Glance"
        with self._user.progress(
            _("Web Calendar Report"), _("Creating Year At A Glance calendar"), 12
        ) as step:
            open_file = self.create_file("fullyearlinked", str(year))

            # page title
            title = self._("%(year)d, At A Glance") % {"year": year}

            # Create page header
            # body has already been added to yearglance  already once
            yearglance, body = self.write_header(nr_up, title, "fullyearlinked", False)

            # create Year Navigation menu
            if self.multiyear and ((self.end_year - self.start_year) > 0):
                body += self.year_navigation(nr_up, str(year))

            # Create Month Navigation Menu
            # identify currentsection for proper highlighting
            body += self.month_navigation(nr_up, year, "fullyearlinked")

            msg = self._(
                "This calendar is meant to give you access "
                "to all your data at a glance compressed into one "
                "page. Clicking on a date will take you to a page "
                "that shows all the events for that date, if there "
                "are any.\n"
            )

            # page description
            content = Html("div", class_="content", id="YearGlance")
            body += content

            content += Html("p", msg, id="description")

            for month in range(1, 13):
                # build the calendar
                monthly_calendar = self.calendar_build(
                    "yg", year, month, clickable=True
                )
                content += monthly_calendar

                # increase progress bar
                step()

            # create blank line for stylesheets
            # write footer section
            footer = self.write_footer(nr_up)
            body += (FULLCLEAR, footer)

            # send calendar page to web output
            # and close the file
            self.xhtmlwriter(yearglance, open_file)

    def one_day(self, event_date, fname_date, day_list):
        """
        This method creates the One Day page for "Year At A Glance"

        event_date -- date for the listed events

        fname_date -- filename date from calendar_build()

        day_list   -- a combination of both dictionaries to be able
                      to create one day
        nyears, date, text, event -- are necessary for figuring the age
        or years married for each year being created...
        """

        nr_up = 1  # number of directory levels up to get to root

        # get year and month from event_date for use in this section
        year = event_date.get_year()
        month = event_date.get_month()

        one_day_file = self.create_file(fname_date, str(year))

        # page title
        title = self._("One Day Within A Year")

        # create page header
        oneday, body = self.write_header(nr_up, title, "OneDay")

        # create Year Navigation menu
        if self.multiyear and ((self.end_year - self.start_year) > 0):
            body += self.year_navigation(nr_up, str(year))

        # Create Month Navigation Menu
        # identify currentsection for proper highlighting
        currentsection = _dd.long_months[month]
        body += self.month_navigation(nr_up, year, currentsection)

        # set date display as in user preferences
        content = Html("div", class_="content", id="OneDay")
        body += content
        evt = fname_date[:8]
        found = (evt, None, None)
        for event in self.event_list:
            if event[0] == evt:
                found = event
                break
        my_title = Html()
        url = "#"
        if found[1] is not None:
            url = event[1] + self.ext
            prevd = Date(int(event[1][:4]), int(event[1][4:6]), int(event[1][6:]))
            my_title = Html("a", "\u276e", href=url, title=self.rlocale.get_date(prevd))
        else:
            my_title = Html("<em>&nbsp;&nbsp;</em>")
        my_title += Html("</a>")
        my_title += "&nbsp;&nbsp;"
        my_title += self.rlocale.date_displayer.display(event_date)
        my_title += "&nbsp;&nbsp;"
        if found[2] is not None:
            url = event[2] + self.ext
            nextd = Date(int(event[2][:4]), int(event[2][4:6]), int(event[2][6:]))
            my_title += Html(
                "a", "\u276f", href=url, title=self.rlocale.get_date(nextd)
            )
        else:
            my_title += Html("<b>&nbsp;&nbsp;</b>")
        content += Html("h3", my_title, inline=True)

        # list the events
        ordered = Html("ol")
        content += ordered
        for (
            dummy_nyears,
            dummy_date,
            text,
            event,
            dummy_age_at_death,
            dummy_dead_event_date,
        ) in day_list:
            ordered += Html(
                "li", text, inline=False if event == "Anniversary" else True
            )

        # create blank line for stylesheets
        # write footer section
        footer = self.write_footer(nr_up)
        body += (FULLCLEAR, footer)

        # send calendar page to web output
        # and close the file
        self.xhtmlwriter(oneday, one_day_file)

    def build_url_fname_html(self, fname, subdir=None, prefix=None):
        """
        build the url for the file name with sub directories and extension
        """
        return self.build_url_fname(fname, subdir, prefix) + self.ext

    def build_url_fname(self, fname, subdir, prefix=None):
        """
        Create part of the URL given the filename and optionally
        the subdirectory. If the subdirectory is given, then two extra levels
        of subdirectory are inserted between 'subdir' and the filename.
        The reason is to prevent directories with too many entries.
        If 'prefix' is set, then is inserted in front of the result.

        The extension is added to the filename as well.

        Notice that we do NOT use os.path.join() because we're creating a URL.
        Imagine we run gramps on Windows (heaven forbits), we don't want to
        see backslashes in the URL.
        """
        if win():
            fname = fname.replace("\\", "/")
        subdirs = self.build_subdirs(subdir, fname)
        return (prefix or "") + "/".join(subdirs + [fname])

    def build_subdirs(self, subdir, fname):
        """
        If subdir is given, then two extra levels of subdirectory are inserted
        between 'subdir' and the filename.
        The reason is to prevent directories with too many entries.

        For example,
        this may return ['ppl', '8', '1'] given 'ppl', "aec934857df74d36618"
        """
        subdirs = []
        if subdir:
            subdirs.append(subdir)
            subdirs.append(fname[-1].lower())
            subdirs.append(fname[-2].lower())
        return subdirs

    def get_name(self, person, maiden_name=None):
        """
        Return person's name, unless maiden_name given, unless married_name
        listed.

        person -- person to get short name from
        maiden_name  -- either a woman's maiden name or man's surname
        """
        # Get all of a person's names:
        primary_name = person.primary_name
        married_name = None
        names = [primary_name] + person.get_alternate_names()
        for name in names:
            if int(name.get_type()) == NameType.MARRIED:
                married_name = name
                break

        # Now, decide which to use:
        if maiden_name is not None:
            if married_name is not None:
                name = Name(married_name)
            else:
                name = Name(primary_name)
                surname_obj = name.get_primary_surname()
                surname_obj.set_surname(maiden_name)
        else:
            name = Name(primary_name)
        name.set_display_as(self.name_format)
        return _nd.display_name(name)

    def collect_data(self, this_year):
        """
        This method runs through the data, and collects the relevant dates
        and text.
        """
        db = self.database

        people = db.iter_person_handles()
        people = self.filter.apply(db, people, user=self._user)

        with self._user.progress(
            _("Web Calendar Report"), _("Reading database..."), len(people)
        ) as step:
            for person in map(db.get_person_from_handle, people):
                step()

                family_list = person.get_family_handle_list()
                birth_ref = person.get_birth_ref()
                birth_date = Date()
                if birth_ref:
                    birth_event = db.get_event_from_handle(birth_ref.ref)
                    birth_date = birth_event.get_date_object()

                death_ref = person.get_death_ref()
                person_death = Date()
                age_at_death = None
                if death_ref and birth_date:
                    death_event = db.get_event_from_handle(death_ref.ref)
                    death_date = death_event.get_date_object()
                    person_death = death_date
                    if birth_date != Date() and birth_date.is_valid() and death_date:
                        age_at_death = death_date - birth_date
                        age_at_death = age_at_death.format(dlocale=self.rlocale)

                # determine birthday information???
                if self.birthday and birth_date != Date() and birth_date.is_valid():
                    birth_date = gregorian(birth_date)

                    year = birth_date.get_year()
                    month = birth_date.get_month()
                    day = birth_date.get_day()

                    # date to figure if someone is still alive
                    # current year of calendar, month nd day is their birth
                    # month and birth day
                    prob_alive_date = Date(this_year, month, day)
                    month, day = birth_date.anniversary(this_year)
                    # add some things to handle maiden name:
                    father_surname = None  # husband, actually
                    if person.gender == Person.FEMALE:
                        # get husband's last name:
                        if self.maiden_name in ["spouse_first", "spouse_last"]:
                            if family_list:
                                if self.maiden_name == "spouse_first":
                                    fhandle = family_list[0]
                                else:
                                    fhandle = family_list[-1]
                                fam = db.get_family_from_handle(fhandle)
                                father_handle = fam.get_father_handle()
                                mother_handle = fam.get_mother_handle()
                                if mother_handle == person.handle:
                                    if father_handle:
                                        father = db.get_person_from_handle(
                                            father_handle
                                        )
                                        if father is not None:
                                            father_surname = _regular_surname(
                                                person.gender, father.get_primary_name()
                                            )
                    short_name = self.get_name(person, father_surname)
                    alive = probably_alive(person, db, prob_alive_date)
                    if (self.alive and alive) or not self.alive:
                        # add link to NarrativeWeb
                        if self.link_to_narweb:
                            if self.narweb_prefix[:1] != "/":
                                prfx = "../" + self.narweb_prefix
                            else:
                                prfx = self.narweb_prefix
                            text = str(
                                Html(
                                    "a",
                                    short_name,
                                    href=self.build_url_fname_html(
                                        person.handle, "ppl", prefix=prfx
                                    ),
                                )
                            )
                        else:
                            text = short_name
                        if age_at_death is None:
                            self.add_day_item(
                                text,
                                year,
                                month,
                                day,
                                "Birthday",
                                age_at_death,
                                birth_date,
                            )
                        else:
                            self.add_day_item(
                                text,
                                year,
                                month,
                                day,
                                "Birthday",
                                age_at_death,
                                person_death,
                            )
                death_event = get_death_or_fallback(db, person)
                if death_event:
                    death_date = death_event.get_date_object()
                else:
                    death_date = None
                # primary_name = person.primary_name
                # name = Name(primary_name)
                if self.death_anniv and death_date:
                    year = death_date.get_year() or this_year
                    month = death_date.get_month()
                    day = death_date.get_day()

                    short_name = self.get_name(person)
                    prob_alive_date = Date(this_year, month, day)
                    alive = probably_alive(person, db, prob_alive_date)
                    if (self.alive and alive) or not self.alive:
                        # add link to NarrativeWeb
                        if self.link_to_narweb:
                            if self.narweb_prefix[:1] != "/":
                                prfx = "../" + self.narweb_prefix
                            else:
                                prfx = self.narweb_prefix
                            navpfx = self.build_url_fname_html(
                                person.handle, "ppl", prefix=prfx
                            )
                            text = str(Html("a", short_name, href=navpfx))
                        else:
                            text = short_name
                        self.add_day_item(
                            text, year, month, day, "Death", age_at_death, death_date
                        )

                # add anniversary if requested
                if self.anniv:
                    for fhandle in family_list:
                        fam = db.get_family_from_handle(fhandle)
                        father_handle = fam.get_father_handle()
                        mother_handle = fam.get_mother_handle()
                        if father_handle == person.handle:
                            spouse_handle = mother_handle
                        else:
                            continue  # with next person if this was
                            # the marriage event
                        if spouse_handle:
                            spouse = db.get_person_from_handle(spouse_handle)
                            if spouse:
                                spouse_name = self.get_name(spouse)
                                short_name = self.get_name(person)
                                death_ref = spouse.get_death_ref()
                                spouse_death = Date()
                                if death_ref:
                                    death_event = db.get_event_from_handle(
                                        death_ref.ref
                                    )
                                    death_date = death_event.get_date_object()
                                    if death_date != Date() and death_date.is_valid():
                                        spouse_death = death_date
                            first_died = Date()
                            if person_death == Date():
                                first_died = spouse_death
                            elif spouse_death != Date():
                                first_died = (
                                    person_death
                                    if spouse_death > person_death
                                    else spouse_death
                                )
                            else:
                                first_died = person_death

                            # will return a marriage event or False if not
                            # married any longer
                            marriage_event = get_marriage_event(db, fam)
                            if marriage_event:
                                event_date = marriage_event.get_date_object()
                                if event_date is not Date() and event_date.is_valid():
                                    event_date = gregorian(event_date)
                                    year = event_date.get_year()
                                    month = event_date.get_month()
                                    day = event_date.get_day()

                                    # date to figure if someone is still alive
                                    prob_alive_date = Date(this_year, month, day)
                                    wedding_age = None
                                    if first_died != Date():
                                        wedding_age = first_died - event_date
                                        wedding_age = wedding_age.format(
                                            dlocale=self.rlocale
                                        )
                                    divorce_event = get_divorce_event(db, fam)
                                    if divorce_event:
                                        d_date = divorce_event.get_date_object()
                                        if d_date is not Date() and d_date.is_valid():
                                            d_date = gregorian(d_date)
                                            if d_date != Date():
                                                w_age = d_date - event_date
                                                w_age = w_age.format(
                                                    dlocale=self.rlocale
                                                )
                                                wedding_age = w_age
                                                first_died = d_date

                                    if self.link_to_narweb:
                                        if self.narweb_prefix[:1] != "/":
                                            prefx = "../" + self.narweb_prefix
                                        else:
                                            prefx = self.narweb_prefix
                                        reference = self.build_url_fname_html(
                                            spouse_handle, "ppl", prefix=prefx
                                        )
                                        spouse_name = str(
                                            Html("a", spouse_name, href=reference)
                                        )
                                        href1 = self.build_url_fname_html(
                                            person.handle, "ppl", prefix=prefx
                                        )
                                        short_name = str(
                                            Html("a", short_name, href=href1)
                                        )

                                    alive1 = probably_alive(person, db, prob_alive_date)
                                    alive2 = probably_alive(spouse, db, prob_alive_date)
                                    if first_died == Date():
                                        first_died = Date(0, 0, 0)
                                    if (
                                        self.alive and (alive1 or alive2)
                                    ) or not self.alive:
                                        spse = self._("%(spouse)s and" " %(person)s")
                                        text = spse % {
                                            "spouse": spouse_name,
                                            "person": short_name,
                                        }

                                        self.add_day_item(
                                            text,
                                            year,
                                            month,
                                            day,
                                            "Anniversary",
                                            wedding_age,
                                            first_died,
                                        )

    def write_footer(self, nr_up):
        """
        Writes the footer section of the pages
        'nr_up' - number of directory levels up, started from current page,
                  to the root of the directory tree (i.e. to self.html_dir).
        """

        # begin calendar footer
        with Html("div", id="footer", role="Footer-End") as footer:
            amsg = None
            if self.author and self.email:
                msg = '<a href="mailto:{email}?subject={subject}">{author}</a>'
                # Translators: This is the email subject line in the Web Calendar
                subject = self._("WebCal")
                amsg = msg.format(email=self.email, subject=subject, author=self.author)
            elif self.author:
                amsg = "%(author)s" % {"author": self.author}

            # Display date as user set in preferences
            date = self.rlocale.date_displayer.display(Today())
            bhtml = '<a href="' + URL_HOMEPAGE + '">'
            msg = self._(
                "Generated by %(gramps_home_html_start)s"
                "Gramps%(html_end)s on %(date)s"
            ) % {"gramps_home_html_start": bhtml, "html_end": "</a>", "date": date}
            copy_nr = self.copy
            if copy_nr == 0:
                if self.author:
                    amsg = "&copy; %s" % amsg
            msg += " (%s)" % amsg
            footer += Html("p", msg, id="createdate")

            text = ""
            if 0 < copy_nr < len(_CC):
                subdirs = [".."] * nr_up
                # Note. We use '/' here because it is a URL,
                # not a OS dependent pathname
                fname = "/".join(subdirs + ["images"] + ["somerights20.gif"])
                text = _CC[copy_nr] % {"gif_fname": fname}

            footer += Html("p", text, id="copyright")

        # return footer to its callers
        return footer

    def xhtmlwriter(self, page, open_file):
        """
        This function is simply to make the web page look pretty and readable
        It is not for the browser, but for us, humans
        """

        # writes the file out from the page variable; Html instance
        # This didn't work for some reason, but it does in NarWeb:
        # page.write(partial(print, file=of.write))
        page.write(lambda line: open_file.write(line + "\n"))
        # close the file now...
        self.close_file(open_file)

    def write_report(self):
        """
        The short method that runs through each month and creates a page.
        """
        # get data from database for birthdays/ anniversaries
        self.collect_data(self.start_year)

        # Copy all files for the calendars being created
        self.copy_calendar_files()

        if self.multiyear:
            # limit number of years to eighteen (18) years and only one row
            # of years
            nyears = (self.end_year - self.start_year) + 1
            num_years = nyears if 0 < nyears < 19 else 18

            for cal_year in range(self.start_year, (self.start_year + num_years)):
                # initialize the holidays dict to fill:
                self.holidays = {}

                # get the information, zero is equal to None
                if self.country != 0:
                    self.__get_holidays(cal_year)

                # create webcalendar() calendar pages
                self.webcalendar(cal_year)

                # create "Year At A Glance" and
                # "One Day" calendar pages
                if self.fullyear:
                    self.year_glance(cal_year)

                self.create_page_index()

        # a single year
        else:
            cal_year = self.start_year

            self.holidays = {}

            # get the information, first from holidays:
            if self.country != 0:
                self.__get_holidays(cal_year)

            # create webcalendar() calendar pages
            self.webcalendar(cal_year)

            # create "Year At A Glance" and
            # "One Day" calendar pages
            if self.fullyear:
                self.year_glance(cal_year)

            if self.link_to_narweb:
                self.create_page_index()

    def create_page_index(self):
        """
        Create the page index called by the narrativeweb.
        """
        output_file = self.create_file("index", "")

        # page title
        title = self.title_text

        nr_up = 0

        # Create page header
        # body has already been added to yearglance  already once
        index, body = self.write_header(nr_up, title, "index", False)

        # create Year Navigation menu
        current_year = time.strftime("%Y", time.gmtime())
        body += self.year_navigation(nr_up, str(current_year))

        # create blank line for stylesheets
        # write footer section
        footer = self.write_footer(nr_up)
        body += (FULLCLEAR, footer)

        # send calendar page to web output
        # and close the file
        self.xhtmlwriter(index, output_file)


# -----------------------------------------------------------------------------
#                             WebCalOptions; Creates the Menu
# ------------------------------------------------------------------------------
class WebCalOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        self.__filter = None
        self.__links = None
        self.__prefix = None
        db_options = name + " " + dbase.get_dbname()
        MenuReportOptions.__init__(self, db_options, dbase)
        self.__multiyear = None
        self.__start_year = None
        self.__end_year = None
        self.__after_year = None

    def add_menu_options(self, menu):
        """
        Add options to the menu for the web calendar.
        """
        self.__add_report_options(menu)
        self.__add_report2_options(menu)
        self.__add_content_options(menu)
        self.__add_advanced_options(menu)
        self.__add_notes_options(menu)

    def __add_report_options(self, menu):
        """
        Options on the "Report Options" tab.
        """
        category_name = _("Report Options")

        dbname = self.__db.get_dbname()
        default_dir = dbname + "_WEBCAL"
        target = DestinationOption(
            _("Destination"),
            os.path.join(config.get("paths.website-directory"), default_dir),
        )
        target.set_help(_("The destination directory for the web files"))
        target.set_directory_entry(True)
        menu.add_option(category_name, "target", target)

        title = StringOption(_("Calendar Title"), _("My Family Calendar"))
        title.set_help(_("The title of the calendar"))
        menu.add_option(category_name, "title", title)

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(
            _("Select filter to restrict people that appear on calendar")
        )
        menu.add_option(category_name, "filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Filter Person"))
        self.__pid.set_help(_("The center person for the filter"))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        self.__update_filters()

        ext = EnumeratedListOption(_("File extension"), ".html")
        for etype in _WEB_EXT:
            ext.add_item(etype, etype)
        ext.set_help(_("The extension to be used for the web files"))
        menu.add_option(category_name, "ext", ext)

        cright = EnumeratedListOption(_("Copyright"), 0)
        for index, copt in enumerate(_COPY_OPTIONS):
            cright.add_item(index, copt)
        cright.set_help(_("The copyright to be used for the web files"))
        menu.add_option(category_name, "cright", cright)

        css_list = sorted(
            [
                (CSS[key]["translation"], CSS[key]["id"])
                for key in list(CSS.keys())
                if CSS[key]["user"]
            ]
        )
        css = EnumeratedListOption(_("StyleSheet"), css_list[0][1])
        for css_item in css_list:
            css.add_item(css_item[1], css_item[0])
        css.set_help(_("The stylesheet to be used for the web pages"))
        menu.add_option(category_name, "css", css)

    def __add_report2_options(self, menu):
        """
        Options on the "Report Options (2)" tab.
        """
        category_name = _("Report Options (2)")

        # We must figure out the value of the first option before we can
        # create the EnumeratedListOption
        fmt_list = _nd.get_name_format()
        defaultnum = _nd.get_default_format()
        default = 0
        for ind, val in enumerate(fmt_list):
            if val[0] == defaultnum:
                default = ind
                break
        name_format = EnumeratedListOption(_("Name format"), fmt_list[default][0])
        for num, name, dummy_fmt_str, dummy_act in fmt_list:
            name_format.add_item(num, name)
        name_format.set_help(_("Select the format to display names"))
        menu.add_option(category_name, "name_format", name_format)

        stdoptions.add_private_data_option(menu, category_name, default=False)

        alive = BooleanOption(_("Include only living people"), True)
        alive.set_help(_("Include only living people in the calendar"))
        menu.add_option(category_name, "alive", alive)

        locale_opt = stdoptions.add_localization_option(menu, category_name)
        stdoptions.add_date_format_option(menu, category_name, locale_opt)

    def __add_content_options(self, menu):
        """
        Options on the "Content Options" tab.
        """
        category_name = _("Content Options")

        # set to today's date for use in menu, etc.
        today = Today()

        self.__multiyear = BooleanOption(_("Create multiple year calendars"), False)
        self.__multiyear.set_help(
            _("Whether to create Multiple year " "calendars or not.")
        )
        menu.add_option(category_name, "multiyear", self.__multiyear)
        self.__multiyear.connect("value-changed", self.__multiyear_changed)

        self.__start_year = NumberOption(
            _("Start Year for the Calendar(s)"), today.get_year(), 1900, 3000
        )
        self.__start_year.set_help(
            _("Enter the starting year for the calendars" " between 1900 - 3000")
        )
        menu.add_option(category_name, "start_year", self.__start_year)

        self.__end_year = NumberOption(
            _("End Year for the Calendar(s)"), today.get_year(), 1900, 3000
        )
        self.__end_year.set_help(
            _("Enter the ending year for the calendars " "between 1900 - 3000.")
        )
        menu.add_option(category_name, "end_year", self.__end_year)

        self.__multiyear_changed()

        country = EnumeratedListOption(_("Country for holidays"), 0)
        holiday_table = libholiday.HolidayTable()
        countries = holiday_table.get_countries()
        countries.sort()
        # if (len(countries) == 0 or
        #        (len(countries) > 0 and countries[0] != '')):
        if not countries or (countries and countries[0] != ""):
            countries.insert(0, "")
        count = 0
        for cntry in countries:
            country.add_item(count, cntry)
            count += 1
        country.set_help(_("Holidays will be included for the selected " "country"))
        menu.add_option(category_name, "country", country)

        # Default selection ????
        start_dow = EnumeratedListOption(_("First day of week"), 1)
        for count in range(1, 8):
            start_dow.add_item(count, _dd.long_days[count].capitalize())
        start_dow.set_help(_("Select the first day of the week " "for the calendar"))
        menu.add_option(category_name, "start_dow", start_dow)

        maiden_name = EnumeratedListOption(_("Birthday surname"), "own")
        maiden_name.add_item(
            "spouse_first",
            _("Wives use husband's surname " "(from first family listed)"),
        )
        maiden_name.add_item(
            "spouse_last", _("Wives use husband's surname " "(from last family listed)")
        )
        maiden_name.add_item("own", _("Wives use their own surname"))
        maiden_name.set_help(_("Select married women's displayed surname"))
        menu.add_option(category_name, "maiden_name", maiden_name)

    def __add_notes_options(self, menu):
        """
        Options on the "Months Notes" tabs.
        """
        category_name = _("Jan - Jun Notes")

        note_jan = NoteOption(_("January Note"))
        note_jan.set_help(_("The note for the month of January"))
        menu.add_option(category_name, "note_jan", note_jan)

        note_feb = NoteOption(_("February Note"))
        note_feb.set_help(_("The note for the month of February"))
        menu.add_option(category_name, "note_feb", note_feb)

        note_mar = NoteOption(_("March Note"))
        note_mar.set_help(_("The note for the month of March"))
        menu.add_option(category_name, "note_mar", note_mar)

        note_apr = NoteOption(_("April Note"))
        note_apr.set_help(_("The note for the month of April"))
        menu.add_option(category_name, "note_apr", note_apr)

        note_may = NoteOption(_("May Note"))
        note_may.set_help(_("The note for the month of May"))
        menu.add_option(category_name, "note_may", note_may)

        note_jun = NoteOption(_("June Note"))
        note_jun.set_help(_("The note for the month of June"))
        menu.add_option(category_name, "note_jun", note_jun)

        category_name = _("Jul - Dec Notes")

        note_jul = NoteOption(_("July Note"))
        note_jul.set_help(_("The note for the month of July"))
        menu.add_option(category_name, "note_jul", note_jul)

        note_aug = NoteOption(_("August Note"))
        note_aug.set_help(_("The note for the month of August"))
        menu.add_option(category_name, "note_aug", note_aug)

        note_sep = NoteOption(_("September Note"))
        note_sep.set_help(_("The note for the month of September"))
        menu.add_option(category_name, "note_sep", note_sep)

        note_oct = NoteOption(_("October Note"))
        note_oct.set_help(_("The note for the month of October"))
        menu.add_option(category_name, "note_oct", note_oct)

        note_nov = NoteOption(_("November Note"))
        note_nov.set_help(_("The note for the month of November"))
        menu.add_option(category_name, "note_nov", note_nov)

        note_dec = NoteOption(_("December Note"))
        note_dec.set_help(_("The note for the month of December"))
        menu.add_option(category_name, "note_dec", note_dec)

    def __add_advanced_options(self, menu):
        """
        Options for the advanced menu
        """

        category_name = _("Advanced Options")

        encoding = EnumeratedListOption(
            _("Character set encoding"), _CHARACTER_SETS[0][1]
        )
        for eopt in _CHARACTER_SETS:
            encoding.add_item(eopt[1], eopt[0])
        encoding.set_help(_("The encoding to be used for the web files"))
        menu.add_option(category_name, "encoding", encoding)

        makeoneday = BooleanOption(
            _("Create one day event pages for" " Year At A Glance calendar"), False
        )
        makeoneday.set_help(_("Whether to create one day pages or not"))
        menu.add_option(category_name, "makeoneday", makeoneday)

        birthdays = BooleanOption(_("Include birthdays"), True)
        birthdays.set_help(_("Include birthdays in the calendar"))
        menu.add_option(category_name, "birthdays", birthdays)

        anniversaries = BooleanOption(_("Include anniversaries"), True)
        anniversaries.set_help(_("Include anniversaries in the calendar"))
        menu.add_option(category_name, "anniversaries", anniversaries)

        anniversaries = BooleanOption(_("Include death dates"), False)
        anniversaries.set_help(_("Include death anniversaries in the calendar"))
        menu.add_option(category_name, "death_anniv", anniversaries)

        self.__links = BooleanOption(_("Link to Narrated Web Report"), False)
        self.__links.set_help(_("Whether to link data to web report or not"))
        menu.add_option(category_name, "link_to_narweb", self.__links)
        self.__links.connect("value-changed", self.__links_changed)

        today = Today()
        default_before = config.get("behavior.max-age-prob-alive")
        self.__after_year = NumberOption(
            _("Show data only after year"),
            (today.get_year() - default_before),
            0,
            today.get_year(),
        )
        self.__after_year.set_help(
            _(
                "Show data only after this year."
                " Default is current year - "
                " 'maximum age probably alive' which is "
                "defined in the dates preference tab."
            )
        )
        menu.add_option(category_name, "after_year", self.__after_year)

        dbname = self.__db.get_dbname()
        default_prefix = "../../" + dbname + "_NAVWEB/"
        self.__prefix = StringOption(_("Link prefix"), default_prefix)
        self.__prefix.set_help(
            _("A Prefix on the links to take you to " "Narrated Web Report")
        )
        menu.add_option(category_name, "prefix", self.__prefix)

        self.__links_changed()

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        filter_list = utils.get_person_filters(person, False)
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if 1 <= filter_value <= 4:
            # Filters 1, 2, 3 and 4 rely on the center person
            self.__pid.set_available(True)
        else:
            # The rest don't
            self.__pid.set_available(False)

    def __multiyear_changed(self):
        """
        Handles the ability to print multiple year calendars or not?
        """
        mgobn = lambda name: self.menu.get_option_by_name(name)
        self.__multiyear = mgobn("multiyear")
        self.__start_year = mgobn("start_year")
        self.__end_year = mgobn("end_year")

        if self.__start_year:
            self.__start_year.set_available(True)
            if self.__multiyear.get_value():
                self.__end_year.set_available(True)
            else:
                self.__end_year.set_available(False)

    def __links_changed(self):
        """
        Handle checkbox change.
        """
        if self.__links.get_value():
            self.__prefix.set_available(True)
        else:
            self.__prefix.set_available(False)


def _regular_surname(sex, name):
    """
    Returns a name string built from the components of the Name instance.
    """
    dummy_gender = sex
    surname = name.get_surname()
    suffix = name.get_suffix()
    if suffix:
        # TODO for Arabic, should the next line's comma be translated?
        surname = surname + ", " + suffix
    return surname


# Simple utility list to convert Gramps day-of-week numbering
# to calendar.firstweekday numbering
DOW_GRAMPS2ISO = [
    -1,
    calendar.SUNDAY,
    calendar.MONDAY,
    calendar.TUESDAY,
    calendar.WEDNESDAY,
    calendar.THURSDAY,
    calendar.FRIDAY,
    calendar.SATURDAY,
]


def get_marriage_event(db, family):
    """
    marriage_event will either be the marriage event or False
    """

    marriage_event = False
    for event_ref in family.get_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event.type.is_marriage():
            marriage_event = event
            break

    # return the marriage event or False to it caller
    return marriage_event


def get_divorce_event(db, family):
    """
    divorce will either be the divorce event or False
    """

    divorce_event = False
    for event_ref in family.get_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event.type.is_divorce():
            divorce_event = event
            break

    # return the divorce event or False to it caller
    return divorce_event


def get_first_day_of_month(year, month):
    """
    Compute the first day to display for this month.
    It can also be a day in the previous month.
    """

    # first day of the month
    current_date = datetime.date(year, month, 1)

    # monthinfo is filled using standard Python library
    # calendar.monthcalendar. It fills a list of 7-day-lists. The first day
    # of the 7-day-list is determined by calendar.firstweekday.
    monthinfo = calendar.monthcalendar(year, month)

    current_ord = current_date.toordinal() - monthinfo[0].count(0)
    return current_date, current_ord, monthinfo


def _has_webpage_extension(url):
    """
    determine if a filename has an extension or not...

    url = filename to be checked
    """
    return any(url.endswith(ext) for ext in _WEB_EXT)


def get_day_list(event_date, holiday_list, bday_anniv_list, rlocale=glocale):
    """
    Will fill day_list and return it to its caller: calendar_build()

    holiday_list -- list of holidays for event_date
    bday_anniv_list -- list of birthdays and anniversaries
        for event_date

    event_date -- date for this day_list

    'day_list' - a combination of both dictionaries to be able
        to create one day nyears, date, text, event --- are
        necessary for figuring the age or years married for
        each day being created...
    rlocale -- the locale to use
    """

    trans_text = rlocale.translation.sgettext
    # initialize day_list
    day_list = []

    ##################################################################
    # birthday/ anniversary on this day
    # Date.EMPTY signifies an incomplete date for an event. See add_day_item()
    bday_anniv_list = [
        (t, e, d, n, x) for t, e, d, n, x in bday_anniv_list if d != Date.EMPTY
    ]

    # number of years have to be at least zero
    bday_anniv_list = [
        (t, e, d, n, x)
        for t, e, d, n, x in bday_anniv_list
        if (event_date.get_year() - d.get_year()) >= 0
    ]

    # a holiday
    # zero will force holidays to be first in list
    nyears = 0

    for text, event, date, notused, notused in holiday_list:
        day_list.append((nyears, date, text, event, notused, notused))

    # birthday and anniversary list
    for text, event, date, age_at_death, dead_event_date in bday_anniv_list:
        # number of years married, ex: 10
        nyears = event_date.get_year() - date.get_year()

        # number of years for birthday, ex: 10 years
        age_str = event_date - date
        # age_str.format(precision=1, as_age=False, dlocale=rlocale)
        age_str = age_str.format(precision=1, as_age=False, dlocale=rlocale)

        symbols = Symbols()
        death_idx = config.get("utf8.death-symbol")
        death_symbol = symbols.get_death_symbol_for_char(death_idx)

        # a birthday
        if event == "Birthday":
            if age_at_death is not None:
                trans_date = trans_text("Died %(death_date)s.")
                translated_date = rlocale.get_date(dead_event_date)
                mess = trans_date % {"death_date": translated_date}
                age = ", <font size='+1' ><b>%s</b></font> <em>%s (%s)" % (
                    death_symbol,
                    mess,
                    age_at_death,
                )
            else:
                # Translators: expands to smth like "12 years old",
                # where "12 years" is already localized to your language
                age = ", <em>"
                date_y = date.get_year()
                trans_date = trans_text("Born %(birth_date)s.")
                old_date = trans_text("%s old")
                translated_date = rlocale.get_date(dead_event_date)
                age += old_date % (
                    str(age_str)
                    if (date_y != 0)
                    else trans_date % {"birth_date": translated_date}
                )
            txt_str = text + age + "</em>"

        # a death
        if event == "Death":
            txt_str = (
                text
                + ", "
                + death_symbol
                + " <em>"
                + (_("%s since death") % str(age_str) if nyears else _("death"))
                + "</em>"
            )

        # an anniversary
        elif event == "Anniversary":
            if nyears == 0:
                txt_str = trans_text("%(couple)s, <em>wedding</em>") % {"couple": text}
            else:
                if age_at_death is not None:
                    age = "%s %s" % (trans_text("Married"), age_at_death)
                    txt_str = "%s, <em>%s" % (text, age)
                    if (
                        isinstance(dead_event_date, Date)
                        and dead_event_date.get_year() > 0
                    ):
                        txt_str += " (" + trans_text("Until") + " "
                        txt_str += rlocale.get_date(dead_event_date)
                        txt_str += ")</em>"
                    else:
                        txt_str += "</em>"
                else:
                    age = "<em>%s" % nyears
                    # Translators: leave all/any {...} untranslated
                    ngettext = rlocale.translation.ngettext
                    txt_str = ngettext(
                        "{couple}, {years} year anniversary",
                        "{couple}, {years} year anniversary",
                        nyears,
                    ).format(couple=text, years=age)
                    txt_str += "</em>"
            txt_str = Html("span", txt_str, class_="yearsmarried")

        day_list.append((nyears, date, txt_str, event, age_at_death, dead_event_date))

    # sort them based on number of years
    # holidays will always be on top of event list
    day_list = sorted(day_list, key=lambda x: (isinstance(x[0], str), x[0]))

    # return to its caller calendar_build()
    return day_list



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..docgen import PaperSize
from ...const import PAPERSIZE

# -------------------------------------------------------------------------
#
# Try to abstract SAX1 from SAX2
#
# -------------------------------------------------------------------------
from xml.sax import make_parser, handler, SAXParseException

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
paper_sizes = []


# -------------------------------------------------------------------------
#
# PageSizeParser
#
# -------------------------------------------------------------------------
class PageSizeParser(handler.ContentHandler):
    """Parses the XML file and builds the list of page sizes"""

    def __init__(self, paper_list):
        handler.ContentHandler.__init__(self)
        self.paper_list = paper_list
        self.locator = None

    def setDocumentLocator(self, locator):
        self.locator = locator

    def startElement(self, tag, attrs):
        if tag == "page":
            name = attrs["name"]
            height = float(attrs["height"])
            width = float(attrs["width"])
            self.paper_list.append(PaperSize(name, height, width))


# -------------------------------------------------------------------------
#
# Parse XML file. If it fails, use the default
#
# -------------------------------------------------------------------------
try:
    parser = make_parser()
    parser.setContentHandler(PageSizeParser(paper_sizes))
    with open(PAPERSIZE) as the_file:
        parser.parse(the_file)
    paper_sizes.append(PaperSize("Custom Size", -1, -1))  # always in English
except (IOError, OSError, SAXParseException):
    paper_sizes = [
        PaperSize("Letter", 27.94, 21.59),
        PaperSize("Legal", 35.56, 21.59),
        PaperSize("A0", 118.9, 84.1),
        PaperSize("A1", 84.1, 59.4),
        PaperSize("A2", 59.4, 42.0),
        PaperSize("A3", 42.0, 29.7),
        PaperSize("A4", 29.7, 21.0),
        PaperSize("A5", 21.0, 14.8),
        PaperSize("B0", 141.4, 100.0),
        PaperSize("B1", 100.0, 70.7),
        PaperSize("B2", 70.7, 50.0),
        PaperSize("B3", 50.0, 35.3),
        PaperSize("B4", 35.3, 25.0),
        PaperSize("B5", 25.0, 17.6),
        PaperSize("B6", 17.6, 12.5),
        PaperSize("B", 43.18, 27.94),
        PaperSize("C", 55.88, 43.18),
        PaperSize("D", 86.36, 55.88),
        PaperSize("E", 111.76, 86.36),
        PaperSize("Custom Size", -1, -1),  # always in English
    ]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2002       Gary Shao
# Copyright (C) 2007       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2012       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import fontscale

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".drawdoc")


# ------------------------------------------------------------------------
#
# DrawDoc
#
# ------------------------------------------------------------------------
class DrawDoc(metaclass=ABCMeta):
    """
    Abstract Interface for graphical document generators. Output formats
    for graphical reports must implement this interface to be used by the
    report system.
    """

    @abstractmethod
    def start_page(self):
        """
        Start a page.
        """

    @abstractmethod
    def end_page(self):
        """
        End a page.
        """

    def get_usable_width(self):
        """
        Return the width of the text area in centimeters. The value is
        the page width less the margins.
        """
        width = self.paper.get_size().get_width()
        right = self.paper.get_right_margin()
        left = self.paper.get_left_margin()
        return width - (right + left)

    def get_usable_height(self):
        """
        Return the height of the text area in centimeters. The value is
        the page height less the margins.
        """
        height = self.paper.get_size().get_height()
        top = self.paper.get_top_margin()
        bottom = self.paper.get_bottom_margin()
        return height - (top + bottom)

    def string_width(self, fontstyle, text):
        "Determine the width need for text in given font"
        return fontscale.string_width(fontstyle, text)

    def string_multiline_width(self, fontstyle, text):
        "Determine the width need for multiline text in given font"
        return fontscale.string_multiline_width(fontstyle, text)

    @abstractmethod
    def draw_path(self, style, path):
        """
        Draw a path.
        """

    @abstractmethod
    def draw_box(self, style, text, x, y, w, h, mark=None):
        """
        :param mark: :class:`.IndexMark` to use for indexing (if supported)
        """

    @abstractmethod
    def draw_text(self, style, text, x1, y1, mark=None):
        """
        :param mark: :class:`.IndexMark` to use for indexing (if supported)
        """

    @abstractmethod
    def center_text(self, style, text, x1, y1, mark=None):
        """
        :param mark: :class:`.IndexMark` to use for indexing (if supported)
        """

    @abstractmethod
    def rotate_text(self, style, text, x, y, angle, mark=None):
        """
        :param mark: :class:`.IndexMark` to use for indexing (if supported)
        """

    @abstractmethod
    def draw_line(self, style, x1, y1, x2, y2):
        """
        Draw a line.
        """



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham, Martin Hawlisch
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Format of commonly used expressions, making use of a cache to not
recompute
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from html import escape

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.symbols import Symbols
from ..lib import EventType
from ..datehandler import get_date
from ..display.name import displayer as name_displayer
from ..display.place import displayer as place_displayer
from .db import get_birth_or_fallback, get_death_or_fallback, get_marriage_or_fallback


# -------------------------------------------------------------------------
#
# FormattingHelper class
#
# -------------------------------------------------------------------------
class FormattingHelper:
    """Format of commonly used expressions, making use of a cache to not
    recompute
    """

    def __init__(self, dbstate, uistate=None):
        self.dbstate = dbstate
        self.uistate = uistate
        self._text_cache = {}
        self._markup_cache = {}
        self.reload_symbols()

    def reload_symbols(self):
        self.symbols = Symbols()
        self.clear_cache()
        if self.uistate and self.uistate.symbols:
            death_idx = self.uistate.death_symbol
            gsfs = self.symbols.get_symbol_for_string
            self.male = gsfs(self.symbols.SYMBOL_MALE)
            self.female = gsfs(self.symbols.SYMBOL_FEMALE)
            self.bth = gsfs(self.symbols.SYMBOL_BIRTH)
            self.marr = gsfs(self.symbols.SYMBOL_MARRIAGE)
            self.dth = self.symbols.get_death_symbol_for_char(death_idx)
        else:
            death_idx = self.symbols.DEATH_SYMBOL_LATIN_CROSS
            gsf = self.symbols.get_symbol_fallback
            self.male = gsf(self.symbols.SYMBOL_MALE)
            self.female = gsf(self.symbols.SYMBOL_FEMALE)
            self.marr = gsf(self.symbols.SYMBOL_MARRIAGE)
            self.bth = gsf(self.symbols.SYMBOL_BIRTH)
            self.dth = self.symbols.get_death_symbol_fallback(death_idx)

    def format_relation(self, family, line_count, use_markup=False):
        """Format a relation between parents of a family"""
        if not family:
            return ""
        if use_markup:
            if family.handle in self._markup_cache:
                if line_count in self._markup_cache[family.handle]:
                    return self._markup_cache[family.handle][line_count]
        else:
            if family.handle in self._text_cache:
                if line_count in self._text_cache[family.handle]:
                    return self._text_cache[family.handle][line_count]

        text = ""
        marriage = get_marriage_or_fallback(self.dbstate.db, family)
        if marriage and use_markup and marriage.get_type() != EventType.MARRIAGE:
            mdate = "<i>%s %s</i>" % (self.marr, escape(get_date(marriage)))
            mplace = "<i>%s</i>" % escape(
                self.get_place_name(marriage.get_place_handle())
            )
            name = "<i>%s</i>" % str(marriage.get_type())
        elif marriage and use_markup:
            mdate = "%s %s" % (self.marr, escape(get_date(marriage)))
            mplace = escape(self.get_place_name(marriage.get_place_handle()))
            name = str(marriage.get_type())
        elif marriage:
            mdate = "%s %s" % (self.marr, get_date(marriage))
            mplace = self.get_place_name(marriage.get_place_handle())
            name = str(marriage.get_type())
        else:
            mdate = ""
            mplace = ""
            name = ""

        if line_count >= 1:
            text += mdate
            text += "\n"
        if line_count >= 2:
            text += name
            text += "\n"
        if line_count >= 3:
            text += mplace
            text += "\n"

        if not text:
            text = str(family.get_relationship())

        if use_markup:
            if not family.handle in self._markup_cache:
                self._markup_cache[family.handle] = {}
            self._markup_cache[family.handle][line_count] = text
        else:
            if not family.handle in self._text_cache:
                self._text_cache[family.handle] = {}
            self._text_cache[family.handle][line_count] = text

        return text

    def get_place_name(self, place_handle):
        """Obtain a place name"""
        text = ""
        if place_handle:
            place = self.dbstate.db.get_place_from_handle(place_handle)
            if place:
                place_title = place_displayer.display(self.dbstate.db, place)
                if place_title != "":
                    if len(place_title) > 25:
                        text = place_title[:24] + "..."
                    else:
                        text = place_title
        return text

    def format_person(self, person, line_count, use_markup=False):
        """fromat how info about a person should be presented"""
        if not person:
            return ""
        if use_markup:
            if person.handle in self._markup_cache:
                if line_count in self._markup_cache[person.handle]:
                    return self._markup_cache[person.handle][line_count]
            name = escape(name_displayer.display(person))
        else:
            if person.handle in self._text_cache:
                if line_count in self._text_cache[person.handle]:
                    return self._text_cache[person.handle][line_count]
            name = name_displayer.display(person)
        text = name
        if line_count >= 3:
            birth = get_birth_or_fallback(self.dbstate.db, person)
            if birth and use_markup and birth.get_type() != EventType.BIRTH:
                bdate = "<i>%s</i>" % escape(get_date(birth))
                bplace = "<i>%s</i>" % escape(
                    self.get_place_name(birth.get_place_handle())
                )
            elif birth and use_markup:
                bdate = escape(get_date(birth))
                bplace = escape(self.get_place_name(birth.get_place_handle()))
            elif birth:
                bdate = get_date(birth)
                bplace = self.get_place_name(birth.get_place_handle())
            else:
                bdate = ""
                bplace = ""
            death = get_death_or_fallback(self.dbstate.db, person)
            if death and use_markup and death.get_type() != EventType.DEATH:
                ddate = "<i>%s</i>" % escape(get_date(death))
                dplace = "<i>%s</i>" % escape(
                    self.get_place_name(death.get_place_handle())
                )
            elif death and use_markup:
                ddate = escape(get_date(death))
                dplace = escape(self.get_place_name(death.get_place_handle()))
            elif death:
                ddate = get_date(death)
                dplace = self.get_place_name(death.get_place_handle())
            else:
                ddate = ""
                dplace = ""

            if line_count < 5:
                text = "%s\n%s %s\n%s %s" % (name, self.bth, bdate, self.dth, ddate)
            else:
                text = "%s\n%s %s\n  %s\n%s %s\n  %s" % (
                    name,
                    self.bth,
                    bdate,
                    bplace,
                    self.dth,
                    ddate,
                    dplace,
                )
        if use_markup:
            if not person.handle in self._markup_cache:
                self._markup_cache[person.handle] = {}
            self._markup_cache[person.handle][line_count] = text
        else:
            if not person.handle in self._text_cache:
                self._text_cache[person.handle] = {}
            self._text_cache[person.handle][line_count] = text
        return text

    def clear_cache(self):
        """clear the cache of kept format strings"""
        self._text_cache = {}
        self._markup_cache = {}


class ImportInfo:
    """
    Class object that can hold information about the import
    """

    def __init__(self, default_info=None):
        """
        Init of the import class.

        This creates the datastructures to hold info
        """
        if default_info is None:
            self.info = {}
        else:
            self.info = default_info

    def info_text(self):
        """
        Construct an info message from the data in the class.
        """
        text = ""
        for key in self.info:
            # Do not fail; can this fail?
            try:
                text += "%s: %s\n" % (key, self.info[key])
            except:
                pass
        return text



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Russian-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Russian parser
#
# -------------------------------------------------------------------------
class DateParserRU(DateParser):
    modifier_to_int = {
        "перед": Date.MOD_BEFORE,
        "по": Date.MOD_BEFORE,
        "до": Date.MOD_BEFORE,
        "после": Date.MOD_AFTER,
        "п.": Date.MOD_AFTER,
        "п": Date.MOD_AFTER,
        "с": Date.MOD_AFTER,
        "ок": Date.MOD_ABOUT,
        "ок.": Date.MOD_ABOUT,
        "около": Date.MOD_ABOUT,
        "примерно": Date.MOD_ABOUT,
        "прим": Date.MOD_ABOUT,
        "прим.": Date.MOD_ABOUT,
        "приблизительно": Date.MOD_ABOUT,
        "приб.": Date.MOD_ABOUT,
        "прибл.": Date.MOD_ABOUT,
        "приб": Date.MOD_ABOUT,
        "прибл": Date.MOD_ABOUT,
        "от": Date.MOD_FROM,
        "к": Date.MOD_TO,
    }

    quality_to_int = {
        "оценено": Date.QUAL_ESTIMATED,
        "оцен.": Date.QUAL_ESTIMATED,
        "оц.": Date.QUAL_ESTIMATED,
        "оцен": Date.QUAL_ESTIMATED,
        "оц": Date.QUAL_ESTIMATED,
        "вычислено": Date.QUAL_CALCULATED,
        "вычисл.": Date.QUAL_CALCULATED,
        "выч.": Date.QUAL_CALCULATED,
        "вычисл": Date.QUAL_CALCULATED,
        "выч": Date.QUAL_CALCULATED,
    }

    bce = ["до нашей эры", "до н. э.", "до н.э.", "до н э", "до нэ"] + DateParser.bce

    def init_strings(self):
        DateParser.init_strings(self)
        DateParser.calendar_to_int.update(
            {
                "персидский": Date.CAL_PERSIAN,
                "п": Date.CAL_PERSIAN,
            }
        )
        _span_1 = ["с", "от"]
        # _span_2 = ['по', 'до'] # <-- clashes with bce parsing :-(
        _span_2 = ["по"]
        _range_1 = ["между", r"меж\.", "меж"]
        _range_2 = ["и"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Russian displayer
#
# -------------------------------------------------------------------------
class DateDisplayRU(DateDisplay):
    """
    Russian language date display class.
    """

    # TODO fix BUG 7064: non-Gregorian calendars wrongly use BCE notation for negative dates
    # not refactoring _bce_str into base class because it'll be gone under #7064
    _bce_str = "%s до н.э."

    display = DateDisplay.display_formatted

    def dd_dformat04(self, date_val, inflect, long_months):
        """
        day month_name year -- for Russian locale
        """
        year = self._slash_year(date_val[2], date_val[3])
        if date_val[0] == 0:
            if date_val[1] == 0:
                return year
            else:
                return self.format_long_month_year(
                    date_val[1], year, inflect, long_months
                )
        elif date_val[1] == 0:  # month is zero but day is not (see 8477)
            return self.display_iso(date_val)
        elif not hasattr(long_months[date_val[1]], "forms"):  # not a Lexeme
            return "{day:d} {long_month} {year}".format(
                day=date_val[0], long_month=long_months[date_val[1]], year=year
            )
        else:
            return "{day:d} {long_month.forms[Р]} {year}".format(
                day=date_val[0], long_month=long_months[date_val[1]], year=year
            )

    def dd_dformat05(self, date_val, inflect, short_months):
        """
        day month_abbreviation year -- for Russian locale
        """
        year = self._slash_year(date_val[2], date_val[3])
        if date_val[0] == 0:
            if date_val[1] == 0:
                return year
            else:
                return self.format_short_month_year(
                    date_val[1], year, inflect, short_months
                )
        elif date_val[1] == 0:  # month is zero but day is not (see 8477)
            return self.display_iso(date_val)
        elif not hasattr(short_months[date_val[1]], "f"):  # not a Lexeme
            return "{day:d} {short_month} {year}".format(
                day=date_val[0], short_month=short_months[date_val[1]], year=year
            )
        else:
            return "{day:d} {short_month.forms[Р]} {year}".format(
                day=date_val[0], short_month=short_months[date_val[1]], year=year
            )


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("ru_RU", "ru", "russian", "Russian", ("%d.%m.%Y",)), DateParserRU, DateDisplayRU
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Show uncollected objects in a window.
"""
# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import weakref
import sys

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
import gc

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.dialog import InfoDialog
from gramps.gui.utils import is_right_click, ProgressMeter
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Leak
#
# -------------------------------------------------------------------------
class Leak(Gramplet):
    """
    Shows uncollected objects.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)

        flags = gc.DEBUG_UNCOLLECTABLE | gc.DEBUG_SAVEALL
        if hasattr(gc, "DEBUG_OBJECTS"):
            flags = flags | gc.DEBUG_OBJECTS
        gc.set_debug(flags)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.top.set_border_width(6)

        self.label = Gtk.Label(halign=Gtk.Align.START)
        self.top.pack_start(self.label, False, False, 6)

        self.scroll = Gtk.ScrolledWindow()
        # add a listview to the scrollable
        self.list = Gtk.TreeView()
        self.list.set_headers_visible(True)
        self.list.connect("button-press-event", self._button_press)
        self.scroll.add(self.list)
        # make a model
        self.model = Gtk.ListStore(int, str, str)
        self.list.set_model(self.model)

        # set the columns
        self.renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Number"), self.renderer, text=0)
        column.set_resizable(True)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.list.append_column(column)
        column = Gtk.TreeViewColumn(_("Referrer"), self.renderer, text=1)
        column.set_resizable(True)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.list.append_column(column)
        column = Gtk.TreeViewColumn(_("Uncollected object"), self.renderer, text=2)
        column.set_resizable(True)
        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.list.append_column(column)
        self.selection = self.list.get_selection()
        self.top.pack_start(self.scroll, True, True, 6)

        bbox = Gtk.ButtonBox()
        apply_button = Gtk.Button(label=_("Refresh"))
        apply_button.connect("clicked", self.apply_clicked)
        bbox.pack_start(apply_button, False, False, 6)
        self.top.pack_start(bbox, False, False, 6)

        self.top.show_all()

        return self.top

    def main(self):
        self.label.set_text(_("Press Refresh to see initial results"))
        self.model.clear()
        # self.display()    # We should only run this on demand

    def _button_press(self, obj, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.referenced_in()
            return True
        elif is_right_click(event):
            self.refers_to()
            return True

    def referenced_in(self):
        model, _iter = self.selection.get_selected()
        if _iter is not None:
            count = model.get_value(_iter, 0)
            gc.collect(2)
            referrers = gc.get_referrers(self.junk[count])
            text = ""
            for referrer in referrers:
                match = ""
                try:
                    if referrer is not self.junk:
                        match = "**** "
                        for indx, junk in enumerate(self.junk):
                            if referrer is junk:
                                match = str(indx) + ": "
                                break
                        match += str(referrer) + "\n"
                except ReferenceError:
                    match += "weakly-referenced object no longer exists %s" % type(
                        referrer
                    )
                except:
                    print(sys.exc_info())
                text += match
            InfoDialog(_("Referrers of %d") % count, text, parent=self.parent)

    def refers_to(self):
        model, _iter = self.selection.get_selected()
        if _iter is not None:
            count = model.get_value(_iter, 0)
            referents = gc.get_referents(self.junk[count])
            text = ""
            for referent in referents:
                match = ""
                try:
                    match = "****: "
                    for indx, junk in enumerate(self.junk):
                        if referent is junk:
                            match = str(indx) + ": "
                            break
                    match += str(referent) + "\n"
                except ReferenceError:
                    match += "%s weakly-referenced object no longer" " exists\n" % type(
                        referent
                    )
                except:
                    print(sys.exc_info())
                text += match
            InfoDialog(_("%d refers to") % count, text, parent=self.parent)

    def display(self):
        try:
            from bsddb3.db import DBError
        except:
            try:
                from berkeleydb.db import DBError
            except:

                class DBError(Exception):
                    """
                    Dummy.
                    """

        self.parent = self.top.get_toplevel()
        progress = ProgressMeter(
            _("Updating display..."), "", parent=self.parent, can_cancel=True
        )
        self.model.clear()
        self.junk = []
        gc.collect(2)
        self.junk = gc.garbage
        self.label.set_text(_("Uncollected Objects: %s") % str(len(self.junk)))
        progress.set_pass(_("Updating display..."), len(self.junk))
        for count in range(0, len(self.junk)):
            if progress.step():
                break
            try:
                refs = []
                referrers = gc.get_referrers(self.junk[count])
                for referrer in referrers:
                    try:
                        if referrer is not self.junk:
                            for indx in range(0, len(self.junk)):
                                if referrer is self.junk[indx]:
                                    refs.append(str(indx) + " ")
                                    break
                    except:
                        print(sys.exc_info())
                if len(refs) > 3:
                    ref = " ".join(refs[0:2]) + "..."
                else:
                    ref = " ".join(refs)
                try:
                    self.model.append((count, ref, str(self.junk[count])))
                except DBError:
                    self.model.append(
                        (count, ref, "db.DB instance at %s" % id(self.junk[count]))
                    )
                except ReferenceError:
                    self.model.append(
                        (
                            count,
                            ref,
                            "weakly-referenced object no longer exists %s"
                            % type(self.junk[count]),
                        )
                    )
                except TypeError:
                    self.model.append(
                        (
                            count,
                            ref,
                            "Object cannot be displayed %s" % type(self.junk[count]),
                        )
                    )
                except:
                    print(sys.exc_info())
            except ReferenceError:
                InfoDialog(
                    _("Reference Error"), "Refresh to correct", parent=self.parent
                )
        progress.close()

    def apply_clicked(self, obj):
        self.display()



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
   This module provides management of all the drag and drop target information
   required by gramps widgets.

   Adding a new drag and drop target.
   ==================================

   To add a new target: add a new _DdType to in _DdTargets.__init__ and
   then add this new type to the list returned from either all_text_types()
   and all_text_targets() or  or all_gramps_targets() and all_gramps_types().

   Usage
   =====

   The module defines a singleton instance of _DdTargets called DdTargets.

   from .ddtargets import DdTargets

   drag_dest_set(Gtk.DestDefaults.ALL,
                 DdTargets.all_targets(),
                 Gdk.DragAction.COPY)

"""

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".DdTargets")

from gi.repository import Gdk
from gi.repository import Gtk


class _DdType:
    """Represents the fields needed by a drag and drop target."""

    _APP_ID_OFFSET = 40  # Starting value of app_ids

    def __init__(self, container, drag_type, target_flags=0, app_id=None):
        """Create a new DdType:

        drag_type: string holding the name of the type.
        target_flags: int value that will be passed to drop target.
        app_id: integer target id passed to drop target.
        """

        self.drag_type = drag_type
        self.atom_drag_type = Gdk.atom_intern(drag_type, False)
        self.target_flags = target_flags
        self.app_id = app_id or self._calculate_id()
        container.insert(self)

    def _calculate_id(self):
        """Return the next available app_id."""

        idval = _DdType._APP_ID_OFFSET
        _DdType._APP_ID_OFFSET += 1
        return idval

    def target(self):
        """
        Return the full target information in the format required by the
        Gtk functions.
        """
        return Gtk.TargetEntry.new(self.drag_type, self.target_flags, self.app_id)

    def target_data(self):
        """
        Return the target information as a list in the format required by
        Gtk3 functions.
        """
        return [self.drag_type, self.target_flags, self.app_id]

    def target_data_atom(self):
        """
        Return the target information as a list in the format required by
        Gtk3 functions.
        """
        return [self.atom_drag_type, self.target_flags, self.app_id]


class _DdTargets:
    """A single class that manages all the drag and drop targets."""

    _instance = None  # Singleton instance

    def __new__(cls):
        """Ensure that we never have more than one instance."""

        if _DdTargets._instance:
            return _DdTargets._instance
        _DdTargets._instance = object.__new__(cls)
        return _DdTargets._instance

    def __init__(self):
        """Set up the drag and drop targets."""

        self._type_map = {}
        self._app_id_map = {}

        self.ADDRESS = _DdType(self, "paddr")
        self.ATTRIBUTE = _DdType(self, "pattr")
        self.CHILDREF = _DdType(self, "childref")
        self.DATA = _DdType(self, "data_tuple")
        self.EVENT = _DdType(self, "pevent")
        self.EVENTREF = _DdType(self, "eventref")
        self.FAMILY_ATTRIBUTE = _DdType(self, "fattr")
        self.FAMILY_EVENT = _DdType(self, "fevent")
        self.LOCATION = _DdType(self, "location")
        self.MEDIAOBJ = _DdType(self, "media")
        self.MEDIAREF = _DdType(self, "mediaref")
        self.NAME = _DdType(self, "name")
        self.NOTE_LINK = _DdType(self, "note-link")
        self.PLACE_LINK = _DdType(self, "place-link")
        self.PLACENAME = _DdType(self, "placename")
        self.PLACEREF = _DdType(self, "placeref")
        self.REPO_LINK = _DdType(self, "repo-link")
        self.REPOREF = _DdType(self, "reporef")
        self.PERSON_LINK = _DdType(self, "person-link")
        self.FAMILY_LINK = _DdType(self, "family-link")
        self.LINK_LIST = _DdType(self, "link-list")
        self.RAW_LIST = _DdType(self, "raw-list")
        self.HANDLE_LIST = _DdType(self, "handle-list")
        self.PERSONREF = _DdType(self, "personref")
        self.SOURCEREF = _DdType(self, "srcref")
        self.SOURCE_LINK = _DdType(self, "source-link")
        self.SRCATTRIBUTE = _DdType(self, "sattr")
        self.URL = _DdType(self, "url")
        self.SURNAME = _DdType(self, "surname")
        self.CITATION_LINK = _DdType(self, "citation-link")

        # List of all types that are used between
        # gramps widgets but should not be exported
        # to non gramps widgets.
        self._all_gramps_types = [
            self.ADDRESS,
            self.ATTRIBUTE,
            self.CHILDREF,
            self.EVENT,
            self.EVENTREF,
            self.LOCATION,
            self.MEDIAOBJ,
            self.MEDIAREF,
            self.NAME,
            self.NOTE_LINK,
            self.PLACE_LINK,
            self.PLACEREF,
            self.PLACENAME,
            self.PERSON_LINK,
            self.FAMILY_LINK,
            self.LINK_LIST,
            self.RAW_LIST,
            self.HANDLE_LIST,
            self.PERSONREF,
            self.REPO_LINK,
            self.REPOREF,
            self.SOURCEREF,
            self.SOURCE_LINK,
            self.SRCATTRIBUTE,
            self.URL,
            self.SURNAME,
            self.CITATION_LINK,
        ]

        self.CHILD = _DdType(self, "child")
        self.SPOUSE = _DdType(self, "spouse")
        self.TEXT_MIME = _DdType(self, "text/plain", 0, 0)
        self.TEXT = _DdType(self, "TEXT", 0, 1)
        self.STRING = _DdType(self, "STRING", 0, 2)
        self.COMPOUND_TEXT = _DdType(self, "COMPOUND_TEXT", 0, 3)
        self.UTF8_STRING = _DdType(self, "UTF8_STRING", 0, 4)
        self.URI_LIST = _DdType(self, "text/uri-list", 0, 5)
        self.APP_ROOT = _DdType(self, "application/x-rootwin-drop", 0, 6)

        # List of all the text types. These are types
        # that can be interpreted as text.
        self._all_text_types = (
            self.UTF8_STRING,
            self.TEXT,
            self.TEXT_MIME,
            self.STRING,
            self.COMPOUND_TEXT,
        )

    def insert(self, dd_type):
        """
        Add a target to the lookup lists. These lists are
        used purely for performance reasons.
        """

        self._type_map[dd_type.drag_type] = dd_type
        self._app_id_map[dd_type.app_id] = dd_type

    def get_dd_type_from_type_name(self, type_name):
        return self._type_map.get(type_name, None)

    def get_dd_type_from_app_id(self, app_id):
        return self._app_id_map.get(app_id, None)

    def is_text_type(self, type_name):
        return type_name in self.all_text_types()

    def all_text(self):
        return self._all_text_types

    def all_text_types(self):
        """
        Return a list of all the type names that could be
        used as the type of a string.
        """

        return tuple([t.drag_type for t in self._all_text_types])

    def is_gramps_type(self, type_name):
        return type_name in self.all_gramps_types()

    def all_gramps_types(self):
        """Return a list of all the type names that are internal to gramps."""

        return tuple([t.drag_type for t in self._all_gramps_types])

    def all_text_targets(self):
        """Return a list of all the targets that could be used for text."""

        return tuple([t.target() for t in self._all_text_types])

    def all_gramps_targets(self):
        """Return a list off the internal gramps targets."""

        return tuple([t.target() for t in self._all_gramps_types])

    def all_targets(self):
        """Return a list of all the known targets."""
        return self.all_gramps_targets() + self.all_text_targets()

    def all_dtype(self):
        """Return all known Ddtype"""
        return [x for x in self._all_gramps_types] + [x for x in self._all_text_types]


# Create the singleton instance.

DdTargets = _DdTargets()


#
# Below here is test code.
#
if __name__ == "__main__":
    print(repr(DdTargets.all_text_types()))
    print(repr(DdTargets.URL))
    print(DdTargets.is_gramps_type("pevent"))



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

This module is used to share variables, enums and functions between all modules

"""

from collections import defaultdict
from hashlib import md5
import re
import logging
from xml.sax.saxutils import escape

from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.db import get_death_or_fallback
from gramps.gen.lib import EventType, Date
from gramps.gen.plug import BasePluginManager
from gramps.plugins.lib.libgedcom import make_gedcom_date, DATE_QUALITY
from gramps.gen.plug.report import utils
from gramps.plugins.lib.libhtml import Html

HAVE_ICU = False
HAVE_ALPHABETICINDEX = False  # separate check as this is only in ICU 4.6+
try:
    from icu import Locale

    HAVE_ICU = True
    try:
        from icu import AlphabeticIndex as icuAlphabeticIndex

        HAVE_ALPHABETICINDEX = True
    except ImportError:
        from gramps.plugins.webreport.alphabeticindex import (
            AlphabeticIndex as localAlphabeticIndex,
        )
except ImportError:
    try:
        from PyICU import Locale

        HAVE_ICU = True
        try:
            from PyICU import AlphabeticIndex as icuAlphabeticIndex

            HAVE_ALPHABETICINDEX = True
        except ImportError:
            from gramps.plugins.webreport.alphabeticindex import (
                AlphabeticIndex as localAlphabeticIndex,
            )
    except ImportError:
        pass

LOG = logging.getLogger(".NarrativeWeb")

# define clear blank line for proper styling
FULLCLEAR = Html("div", class_="fullclear", inline=True)
# define all possible web page filename extensions
_WEB_EXT = [".html", ".htm", ".shtml", ".php", ".cgi"]
# used to select secured web site or not
HTTP = "http://"
HTTPS = "https://"

GOOGLE_MAPS = "https://maps.googleapis.com/maps/"
# javascript code for marker path
MARKER_PATH = """
  var marker_png = '%s';
"""

# javascript code for Google's FamilyLinks...
FAMILYLINKS = """
  var tracelife = %s;

  window.addEventListener("load", function() {
    var myLatLng = new google.maps.LatLng(%s, %s);

    var mapOptions = {
      scaleControl:    true,
      panControl:      true,
      backgroundColor: '#000000',
      zoom:            %d,
      center:          myLatLng,
      mapTypeId:       google.maps.MapTypeId.ROADMAP
    };
    var map = new google.maps.Map(document.getElementById("map_canvas"),
                                  mapOptions);

    var flightPath = new google.maps.Polyline({
      path:          tracelife,
      strokeColor:   "#FF0000",
      strokeOpacity: 1.0,
      strokeWeight:  2
    });

   flightPath.setMap(map);
  });
"""

# javascript for Google's Drop Markers...
DROPMASTERS = """
  var markers = [];
  var iterator = 0;

  var tracelife = %s;
  var map;
  var myLatLng = new google.maps.LatLng(%s, %s);

  window.addEventListener("load", function() {
    var mapOptions = {
      scaleControl: true,
      zoomControl:  true,
      zoom:         %d,
      mapTypeId:    google.maps.MapTypeId.ROADMAP,
      center:       myLatLng,
    };
    map = new google.maps.Map(document.getElementById("map_canvas"),
                              mapOptions);
  });

  function drop() {
    for (var i = 0; i < tracelife.length; i++) {
      setTimeout(function() {
        addMarker();
      }, i * 1000);
    }
  }

  function addMarker() {
    var location = tracelife[iterator];
    var myLatLng = new google.maps.LatLng(location[1], location[2]);
    var infoWindow = new google.maps.InfoWindow;

    var marker = new google.maps.Marker({
      position:  myLatLng,
      map:       map,
      draggable: true,
      title:     location[0],
      animation: google.maps.Animation.DROP
    });
    markers.push(marker);
    iterator++;
    var title = "<h2>" + location[0] + "</h2>"
    bindInfoWindow(marker, map, infoWindow, title+location[4]);
  }
  function bindInfoWindow(marker, map, infoWindow, html) {
          google.maps.event.addListener(marker, 'click', function() {
              infoWindow.setContent(html);
              infoWindow.open(map, marker);
          });
  }
"""

# javascript for Google's Markers...
MARKERS = """
  var tracelife = %s;
  var map;
  var myLatLng = new google.maps.LatLng(%s, %s);

  window.addEventListener("load", function() {
    var mapOptions = {
      scaleControl:    true,
      panControl:      true,
      backgroundColor: '#000000',
      zoom:            %d,
      center:          myLatLng,
      mapTypeId:       google.maps.MapTypeId.ROADMAP
    };
    map = new google.maps.Map(document.getElementById("map_canvas"),
                              mapOptions);
    addMarkers();
  });

  function addMarkers() {
    var bounds = new google.maps.LatLngBounds();
    var infoWindow = new google.maps.InfoWindow;

    for (var i = 0; i < tracelife.length; i++) {
      var location = tracelife[i];
      var myLatLng = new google.maps.LatLng(location[1], location[2]);

      var marker = new google.maps.Marker({
        position:  myLatLng,
        draggable: true,
        title:     location[0],
        map:       map,
        zIndex:    location[3]
      });
      var title = "<h2>" + location[0] + "</h2>"
      bindInfoWindow(marker, map, infoWindow, title+location[4]);
      bounds.extend(myLatLng);
      if ( i > 1 ) { map.fitBounds(bounds); };
    }
  }
  function bindInfoWindow(marker, map, infoWindow, html) {
          google.maps.event.addListener(marker, 'click', function() {
              infoWindow.setContent(html);
              infoWindow.open(map, marker);
          });
  }
"""

# javascript for OpenStreetMap's markers...
"""
https://openlayers.org/en/latest/examples/
"""

OSM_MARKERS = """
  window.addEventListener("load", function() {
    var map;
    var tracelife = %s;
    var iconStyle = new ol.style.Style({
      image: new ol.style.Icon(({
        anchor: [0.2, 48],
        anchorXUnits: 'fraction',
        anchorYUnits: 'pixels',
        opacity: 1.0,
        src: marker_png
      }))
    });
    var markerSource = new ol.source.Vector({
    });
    for (var i = 0; i < tracelife.length; i++) {
      var loc = tracelife[i];
      var iconFeature = new ol.Feature({
       geometry: new ol.geom.Point(ol.proj.transform([loc[0], loc[1]],
                                                     'EPSG:4326', 'EPSG:3857')),
       name: loc[2],
       data: loc[3],
      });
      iconFeature.setStyle(iconStyle);
      markerSource.addFeature(iconFeature);
    }
    markerLayer = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    tooltip = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    var centerCoord = new ol.proj.transform([%s, %s], 'EPSG:4326', 'EPSG:3857');
    map = new ol.Map({
                 target: 'map_canvas',
                 layers: [new ol.layer.Tile({ source: new ol.source.OSM() }),
                          markerLayer, tooltip],
                 view: new ol.View({ center: centerCoord, zoom: %d })
                 });
"""

STAMEN_MARKERS = """
  window.addEventListener("load", function() {
    var map;
    var tracelife = %s;
    var layer = '%s';
    var iconStyle = new ol.style.Style({
      image: new ol.style.Icon(({
        anchor: [0.2, 48],
        anchorXUnits: 'fraction',
        anchorYUnits: 'pixels',
        opacity: 1.0,
        src: marker_png
      }))
    });
    var markerSource = new ol.source.Vector({
    });
    for (var i = 0; i < tracelife.length; i++) {
      var loc = tracelife[i];
      var iconFeature = new ol.Feature({
       geometry: new ol.geom.Point(ol.proj.transform([loc[0], loc[1]],
                                                     'EPSG:4326', 'EPSG:3857')),
       name: loc[2],
       data: loc[3],
      });
      iconFeature.setStyle(iconStyle);
      markerSource.addFeature(iconFeature);
    }
    var centerCoord = new ol.proj.transform([%s, %s], 'EPSG:4326', 'EPSG:3857');
    markerLayer = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    tooltip = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    map = new ol.Map({
                 target: 'map_canvas',
                 layers: [
                   new ol.layer.Tile({ source: new ol.source.Stamen({
                                         layer: layer }) }),
                          markerLayer, tooltip],
                 view: new ol.View({ center: centerCoord, zoom: %d })
                 });
"""

OPENLAYER = """
    var element = document.getElementById('popup');
    var content = document.getElementById('popup-content');
    var ptitle = document.getElementById('popup-title');
    var closer = document.getElementById('popup-closer');
    var tip = document.getElementById('tooltip');
    var tipcontent = document.getElementById('tooltip-content');
    var tooltip = new ol.Overlay({
      element: tip,
      positioning: 'bottom-center',
      offset: [10, 0],
    });
    map.addOverlay(tooltip);
    var popup = new ol.Overlay({
      element: element,
      positioning: 'bottom-center',
      autoPan: true,
      autoPanAnimation: { duration: 500 },
      stopEvent: true
    });
    map.addOverlay(popup);
    /**
     * Add a click handler to hide the popup.
     * @return {boolean} Don't follow the href.
     */
    closer.onclick = function() {
      popup.setPosition(undefined);
      closer.blur();
      return false;
    };
    map.on('pointermove', function(evt) {
      evt.preventDefault()
      var feature = this.forEachFeatureAtPixel(evt.pixel,
                                               function(feature, layer) {
        return feature;
      });
      map.getTargetElement().style.cursor = feature ? 'pointer' : '';
      if (evt.dragging) {
        tooltip.setPosition(undefined);
        return;
      }
      if (feature) {
        var coordinate = evt.coordinate;
        tipcontent.innerHTML = feature.get('name');
        tooltip.setPosition(coordinate);
      } else {
        tooltip.setPosition(undefined);
      }
    });
    map.on('singleclick', function(evt) {
      evt.preventDefault();
      var feature = map.forEachFeatureAtPixel(evt.pixel,
                                              function(feature, layer) {
        return feature;
      });
      if (feature) {
      var coordinate = evt.coordinate;
      var title = '<h2>' + feature.get('name') + '</h2>';
      ptitle.innerHTML = title;
      content.innerHTML = feature.get('data');
      popup.setPosition(coordinate);
      }
      });
  });
"""

# variables for alphabet_navigation()
_KEYPERSON, _KEYPLACE, _KEYEVENT, _ALPHAEVENT = 0, 1, 2, 3

COLLATE_LANG = glocale.collation

_NAME_STYLE_SHORT = 2
_NAME_STYLE_DEFAULT = 1
_NAME_STYLE_FIRST = 0
_NAME_STYLE_SPECIAL = None

PLUGMAN = BasePluginManager.get_instance()
CSS = PLUGMAN.process_plugin_data("WEBSTUFF")

# _NAME_COL = 3

_WRONGMEDIAPATH = []

_HTML_DBL_QUOTES = re.compile(r'([^"]*) " ([^"]*) " (.*)', re.VERBOSE)
_HTML_SNG_QUOTES = re.compile(r"([^']*) ' ([^']*) ' (.*)", re.VERBOSE)

# Events that are usually a family event
_EVENTMAP = set(
    [
        EventType.MARRIAGE,
        EventType.MARR_ALT,
        EventType.MARR_SETTL,
        EventType.MARR_LIC,
        EventType.MARR_CONTR,
        EventType.MARR_BANNS,
        EventType.ENGAGEMENT,
        EventType.DIVORCE,
        EventType.DIV_FILING,
    ]
)

# Names for stylesheets
_NARRATIVESCREEN = "narrative-screen.css"
_NARRATIVEPRINT = "narrative-print.css"

HOLIDAYS_ASSOC = [
    # LANG INDEX HOLIDAY_NAME
    ("be", 0, "Bulgaria"),
    ("ca", 1, "Canada"),
    ("cs", 2, "Czech Republic"),
    ("cl", 3, "Chile"),
    ("zh", 4, "China"),
    ("hr", 5, "Croatia"),
    ("en", 13, "United States of America"),  # must be here. should be en_US
    ("en_GB", 6, "England"),
    ("fi", 7, "Finland"),
    ("fr_FR", 8, "France"),
    ("de", 9, "Germany"),
    ("ja", 10, "Japan"),
    ("sk", 11, "Slovakia"),
    ("sv", 12, "Sweden"),
    ("he", 14, "Jewish Holidays"),
    ("en_NZ", 15, "New Zealand"),
    ("uk", 16, "Ukraine"),
    ("sr_RS", 17, "Serbia"),
    ("sr_RS@latin", 18, "Serbia (Latin)"),
]


def do_we_have_holidays(lang):
    """
    Associate an index for the holidays depending on the LANG
    """
    for lng, idx, dummy_name in HOLIDAYS_ASSOC:
        if lng == lang:  # i.e. en_US
            return idx
        if lng[:2] == lang:  # i.e. en_US[:2] => en
            return idx
    return None


def get_surname_from_person(dbase, person):
    """
    get the person's surname
    get the primary name
    if group as get the group_as surname
    else get the primary surname of the primary name
         and correct for [global] group_as name
    correct for surnames that are space or None
    """
    primary_name = person.get_primary_name()

    if primary_name.group_as:
        surname = primary_name.group_as
    else:
        group_map = _nd.primary_surname(primary_name)
        surname = dbase.get_name_group_mapping(group_map)

    # Treat people who have no name with those whose name is just
    # 'whitespace'
    if surname is None or surname.isspace():
        surname = ""
    return surname


def sort_people(dbase, handle_list, rlocale=glocale):
    """
    will sort the database people by surname
    @param: dbase           -- The instance of the database
    @param: handle_list     -- The list of handles of people to sort
    @param: rlocale         -- The locale related to the language used for the
                               sort
    @result:                -- A list sorted by surname, each element of which
                               consists of a tuple of (surname, list of handles)
                               where the list of handles is sorted by
                               primary surname, first name, suffix.
                               Surname uses group_as, but primary surname
                               does not.
    get the primary name
    if group as get the group_as surname
    else get the primary surname of the primary name
         and correct for [global] group_as name
    correct for surnames that are space or None
    for each surname sort handles by the surname, first name and suffix
    construct a list of surnames and list of handles
    """
    sname_sub = defaultdict(list)
    sortnames = {}

    for person_handle in handle_list:
        person = dbase.get_person_from_handle(person_handle)
        surname = get_surname_from_person(dbase, person)
        sortnames[person_handle] = _nd.sort_string(person.get_primary_name())
        sname_sub[surname].append(person_handle)

    sorted_lists = []
    temp_list = sorted(sname_sub, key=rlocale.sort_key)

    for name in temp_list:
        if isinstance(name, bytes):
            name = name.decode("utf-8")
        slist = sorted(
            ((sortnames[x], x) for x in sname_sub[name]),
            key=lambda x: rlocale.sort_key(x[0]),
        )
        entries = [x[1] for x in slist]
        sorted_lists.append((name, entries))

    return sorted_lists


def sort_places(dbase, handle_list, rlocale=glocale):
    """
    will sort the database places
    """
    pname_sub = defaultdict(list)

    for place_name in handle_list.keys():
        cname = sname = None
        if len(handle_list[place_name]) == 6:
            (hdle, pname, cname, sname, dummy_id, event) = handle_list[place_name]
        else:
            continue
        place = dbase.get_place_from_handle(hdle)
        pname = _pd.display(dbase, place, fmt=0)
        apname = _pd.display_event(dbase, event, fmt=0)

        pname_sub[pname].append(hdle)
        pname_sub[apname].append((hdle, pname, cname, sname))

    sorted_lists = []
    temp_list = sorted(pname_sub, key=rlocale.sort_key)

    for name in temp_list:
        if isinstance(name, bytes):
            name = name.decode("utf-8")
        sorted_lists.append((name, pname_sub[name][0]))

    return sorted_lists


def sort_event_types(dbase, event_types, event_handle_list, rlocale):
    """
    sort a list of event types and their associated event handles

    @param: dbase -- report database
    @param: event_types -- a dict of event types
    @param: event_handle_list -- all event handles in this database
    """
    rts = rlocale.translation.sgettext
    event_dict = dict((rts(evt_type), list()) for evt_type in event_types)

    for event_handle in event_handle_list:
        event = dbase.get_event_from_handle(event_handle)
        event_type = rts(event.get_type().xml_str())

        # add (gramps_id, date, handle) from this event
        if event_type in event_dict:
            sort_value = event.get_date_object().get_sort_value()
            event_dict[event_type].append((sort_value, event_handle))

    for tup_list in event_dict.values():
        tup_list.sort()

    # return a list of sorted tuples, one per event
    retval = [
        (event_type, event_list) for (event_type, event_list) in event_dict.items()
    ]
    retval.sort(key=lambda item: str(item[0]))

    return retval


# Modified _get_regular_surname from WebCal.py to get prefix, first name,
# and suffix
def _get_short_name(gender, name):
    """Will get suffix for all people passed through it"""

    dummy_gender = gender
    short_name = name.get_first_name()
    suffix = name.get_suffix()
    if suffix:
        # TODO for Arabic, should the next line's comma be translated?
        short_name = short_name + ", " + suffix
    return short_name


def __get_person_keyname(dbase, handle):
    """...."""

    person = dbase.get_person_from_handle(handle)
    return _nd.sort_string(person.get_primary_name())


def __get_place_keyname(dbase, handle):
    """..."""

    return utils.place_name(dbase, handle)


if HAVE_ALPHABETICINDEX:

    class AlphabeticIndex(icuAlphabeticIndex):
        """
        Call the ICU AlphabeticIndex, passing the ICU Locale
        """

        def __init__(self, rlocale):
            self.iculocale = Locale(rlocale.collation)
            super().__init__(self.iculocale)

            # set the maximum number of buckets, the undocumented default is 99
            # Latin + Greek + Cyrillic + Hebrew + Arabic + Tamil + Hiragana +
            # CJK Unified is about 206 different buckets
            self.maxLabelCount = 500  # pylint: disable=invalid-name

            # Add bucket labels for scripts other than the one for the output
            # which is being generated
            self.iculocale.addLikelySubtags()
            default_script = self.iculocale.getDisplayScript()
            used_scripts = [default_script]

            for lang_code in glocale.get_language_dict().values():
                loc = Locale(lang_code)
                loc.addLikelySubtags()
                script = loc.getDisplayScript()
                if script not in used_scripts:
                    used_scripts.append(script)
                    super().addLabels(loc)

else:
    AlphabeticIndex = localAlphabeticIndex


def alphabet_navigation(sorted_alpha_index, rlocale=glocale, rtl=False):
    """
    Will create the alphabet navigation bar for classes IndividualListPage,
    SurnameListPage, PlaceListPage, and EventList

    @param: index_list -- a dictionary of either letters or words
    @param: rlocale    -- The locale to use
    """
    # if no letters, return None to its callers
    if not sorted_alpha_index:
        return None

    num_ltrs = len(sorted_alpha_index)
    num_of_cols = 26
    num_of_rows = (num_ltrs // num_of_cols) + 1

    # begin alphabet navigation division
    with Html("div", id=rtl) as alphabetnavigation:
        index = 0
        output = []
        dup_index = 0
        for dummy_row in range(num_of_rows):
            unordered = Html("ul", class_=rtl)

            cols = 0
            while cols <= num_of_cols and index < num_ltrs:
                menu_item = sorted_alpha_index[index]
                if menu_item == " ":
                    menu_item = "&nbsp;"
                # adding title to hyperlink menu for screen readers and
                # braille writers
                title_txt = "Alphabet Menu: %s" % menu_item
                title_str = rlocale.translation.sgettext(title_txt)
                # deal with multiple ellipsis which are generated for overflow,
                # underflow and inflow labels
                link = menu_item
                if menu_item in output:
                    link = "%s (%i)" % (menu_item, dup_index)
                    dup_index += 1
                output.append(menu_item)

                hyper = Html("a", menu_item, title=title_str, href="#%s" % link)
                unordered.extend(Html("li", hyper, inline=True))

                index += 1
                cols += 1
            num_of_rows -= 1

            alphabetnavigation += unordered

    return alphabetnavigation


def _has_webpage_extension(url):
    """
    determine if a filename has an extension or not...

    @param: url -- filename to be checked
    """
    return any(url.endswith(ext) for ext in _WEB_EXT)


def add_birthdate(dbase, ppl_handle_list, rlocale):
    """
    This will sort a list of child handles in birth order
    For each entry in the list, we'll have :
         birth date
         The transtated birth date for the configured locale
         The transtated death date for the configured locale
         The handle for the child

    @param: dbase           -- The database to use
    @param: ppl_handle_list -- the handle for the people
    @param: rlocale         -- the locale for date translation
    """
    sortable_individuals = []
    for person_handle in ppl_handle_list:
        birth_date = 0  # dummy value in case none is found
        person = dbase.get_person_from_handle(person_handle)
        if person:
            birth_ref = person.get_birth_ref()
            birth1 = ""
            if birth_ref:
                birth = dbase.get_event_from_handle(birth_ref.ref)
                if birth:
                    birth1 = rlocale.get_date(birth.get_date_object())
                    birth_date = birth.get_date_object().get_sort_value()
            death_event = get_death_or_fallback(dbase, person)
            if death_event:
                death = rlocale.get_date(death_event.get_date_object())
            else:
                death = ""
        sortable_individuals.append((birth_date, birth1, death, person_handle))

    # return a list of handles with the individual's birthdate attached
    return sortable_individuals


def _find_birth_date(dbase, individual):
    """
    will look for a birth date within the person's events

    @param: dbase      -- The database to use
    @param: individual -- The individual for who we want to find the birth date
    """
    date_out = None
    birth_ref = individual.get_birth_ref()
    if birth_ref:
        birth = dbase.get_event_from_handle(birth_ref.ref)
        if birth:
            date_out = birth.get_date_object()
            date_out.fallback = False
    else:
        person_evt_ref_list = individual.get_primary_event_ref_list()
        if person_evt_ref_list:
            for evt_ref in person_evt_ref_list:
                event = dbase.get_event_from_handle(evt_ref.ref)
                if event:
                    if event.get_type().is_birth_fallback():
                        date_out = event.get_date_object()
                        date_out.fallback = True
                        LOG.debug("setting fallback to true for '%s'", event)
                        break
    return date_out


def _find_death_date(dbase, individual):
    """
    will look for a death date within a person's events

    @param: dbase      -- The database to use
    @param: individual -- The individual for who we want to find the death date
    """
    date_out = None
    death_ref = individual.get_death_ref()
    if death_ref:
        death = dbase.get_event_from_handle(death_ref.ref)
        if death:
            date_out = death.get_date_object()
            date_out.fallback = False
    else:
        person_evt_ref_list = individual.get_primary_event_ref_list()
        if person_evt_ref_list:
            for evt_ref in person_evt_ref_list:
                event = dbase.get_event_from_handle(evt_ref.ref)
                if event:
                    if event.get_type().is_death_fallback():
                        date_out = event.get_date_object()
                        date_out.fallback = True
                        LOG.debug("setting fallback to true for '%s'", event)
                        break
    return date_out


def build_event_data_by_individuals(dbase, ppl_handle_list):
    """
    creates a list of event handles and event types for this database

    @param: dbase           -- The database to use
    @param: ppl_handle_list -- the handle for the people
    """
    event_handle_list = []
    event_types = []

    for person_handle in ppl_handle_list:
        person = dbase.get_person_from_handle(person_handle)
        if person:
            evt_ref_list = person.get_event_ref_list()
            if evt_ref_list:
                for evt_ref in evt_ref_list:
                    event = dbase.get_event_from_handle(evt_ref.ref)
                    if event:
                        event_types.append(str(event.get_type()))
                        event_handle_list.append(evt_ref.ref)

            person_family_handle_list = person.get_family_handle_list()
            if person_family_handle_list:
                for family_handle in person_family_handle_list:
                    family = dbase.get_family_from_handle(family_handle)
                    if family:
                        family_evt_ref_list = family.get_event_ref_list()
                        if family_evt_ref_list:
                            for evt_ref in family_evt_ref_list:
                                event = dbase.get_event_from_handle(evt_ref.ref)
                                if event:
                                    event_types.append(str(event.type))
                                    event_handle_list.append(evt_ref.ref)

    # return event_handle_list and event types to its caller
    return event_handle_list, event_types


def name_to_md5(text):
    """This creates an MD5 hex string to be used as filename."""

    return md5(text.encode("utf-8")).hexdigest()


def get_gendex_data(database, event_ref):
    """
    Given an event, return the date and place a strings

    @param: database  -- The database
    @param: event_ref -- The event reference
    """
    doe = ""  # date of event
    poe = ""  # place of event
    if event_ref and event_ref.ref:
        event = database.get_event_from_handle(event_ref.ref)
        if event:
            date = event.get_date_object()
            doe = format_date(date)
            if event.get_place_handle():
                place_handle = event.get_place_handle()
                if place_handle:
                    place = database.get_place_from_handle(place_handle)
                    if place:
                        poe = _pd.display(database, place, date)
    return doe, poe


def format_date(date):
    """
    Format the date
    """
    start = date.get_start_date()
    if start != Date.EMPTY:
        cal = date.get_calendar()
        mod = date.get_modifier()
        quality = date.get_quality()
        if quality in DATE_QUALITY:
            qual_text = DATE_QUALITY[quality] + " "
        else:
            qual_text = ""
        if mod == Date.MOD_SPAN:
            val = "%sFROM %s TO %s" % (
                qual_text,
                make_gedcom_date(start, cal, mod, None),
                make_gedcom_date(date.get_stop_date(), cal, mod, None),
            )
        elif mod == Date.MOD_RANGE:
            val = "%sBET %s AND %s" % (
                qual_text,
                make_gedcom_date(start, cal, mod, None),
                make_gedcom_date(date.get_stop_date(), cal, mod, None),
            )
        else:
            val = make_gedcom_date(start, cal, mod, quality)
        return val
    return ""


# This command then defines the 'html_escape' option for escaping
# special characters for presentation in HTML based on the above list.
def html_escape(text):
    """Convert the text and replace some characters with a &# variant."""

    # First single characters, no quotes
    text = escape(text)

    # Deal with double quotes.
    match = _HTML_DBL_QUOTES.match(text)
    while match:
        text = "%s" "&#8220;" "%s" "&#8221;" "%s" % match.groups()
        match = _HTML_DBL_QUOTES.match(text)
    # Replace remaining double quotes.
    text = text.replace('"', "&#34;")

    # Deal with single quotes.
    text = text.replace("'s ", "&#8217;s ")
    match = _HTML_SNG_QUOTES.match(text)
    while match:
        text = "%s" "&#8216;" "%s" "&#8217;" "%s" % match.groups()
        match = _HTML_SNG_QUOTES.match(text)
    # Replace remaining single quotes.
    text = text.replace("'", "&#39;")

    return text



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide a python evaluation window
"""
# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------
import sys
from io import StringIO
import traceback

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PythonEvaluation
#
# -------------------------------------------------------------------------
class PythonEvaluation(Gramplet):
    """
    Allows the user to evaluate python code.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.top.set_border_width(6)

        self.ebuf = self.__add_text_view(_("Evaluation"))
        self.dbuf = self.__add_text_view(_("Output"))
        self.error = self.__add_text_view(_("Error"))

        bbox = Gtk.ButtonBox()
        apply_button = Gtk.Button(label=_("Apply"))
        apply_button.connect("clicked", self.apply_clicked)
        bbox.pack_start(apply_button, False, False, 6)
        clear_button = Gtk.Button(label=_("Clear"))
        clear_button.connect("clicked", self.clear_clicked)
        bbox.pack_start(clear_button, False, False, 6)
        self.top.pack_start(bbox, False, False, 6)

        self.top.show_all()

        return self.top

    def __add_text_view(self, name):
        """
        Add a text view to the interface.
        """
        label = Gtk.Label(halign=Gtk.Align.START)
        label.set_markup("<b>%s</b>" % name)
        self.top.pack_start(label, False, False, 6)
        swin = Gtk.ScrolledWindow()
        swin.set_shadow_type(Gtk.ShadowType.IN)
        tview = Gtk.TextView()
        swin.add(tview)
        self.top.pack_start(swin, True, True, 6)
        return tview.get_buffer()

    def apply_clicked(self, obj):
        text = str(
            self.ebuf.get_text(
                self.ebuf.get_start_iter(), self.ebuf.get_end_iter(), False
            )
        )

        oldout = sys.stdout
        olderr = sys.stderr
        outtext = StringIO()
        errtext = StringIO()
        sys.stdout = outtext
        sys.stderr = errtext
        try:
            exec(text)
        except:
            traceback.print_exc()
        self.dbuf.set_text(outtext.getvalue())
        self.error.set_text(errtext.getvalue())
        sys.stdout = oldout
        sys.stderr = olderr

    def clear_clicked(self, obj):
        self.dbuf.set_text("")
        self.ebuf.set_text("")
        self.error.set_text("")
